/*!
  @file BlockIP.h 
  @brief Definition of BlockIP
*/

//****************************************************************************
/*! 
 \mainpage 
 
BlockIP implements a specialized primal-dual long-step path-following
interior point algorithm for linear, separable convex quadratic, or
separable convex nonlinear problems with primal block structure.

Defining f(x) as  either 

\verbatim
          f(x)= c_1 x_1 + ...+ c_k x_k + c_{k+1} x_{k+1} +
	        1/2 (x_1...x_k x_{k+1}) Q (x_1...x_k x_{k+1})
\endverbatim

or 

\verbatim
          f(x)= f(x_1, ..., x_k,x_{k+1})
\endverbatim

the problem is \n

\verbatim
      min  f(x) 
      subj. to
         N_i x_i = b_i  i=1,...,k                       Block equations 
         (L_1*x_1 + ...+ L_k*x_k) + x_{k+1} = b_{k+1}   Linking constraints 
         0 <= x_i <= u_i   i=1,...,k+1                  Bounds 
\endverbatim

  where x_i i=1..k are the variables for each block
  and x_{k+1} are the slacks of the linking constraints

The normal equations for the dual direction is computed in two parts.
The part associated to blocks is solved through k Cholesky factorizations.
The part associated to linking constraints is computed with a PCG
using a power series preconditioner. Cholesky factorizations are currently 
performed with Ng-Peyton's sprsblkllt; code is ready for others solvers.

A detailed description of BlockIP can be found in

-     J. Castro, Interior-point solver for convex separable block-angular problems,
      Research Report DR 2014/03, Dept. of Statistics and Operations Research, 
      Universitat Politecnica de Catalunya, 2014 (submitted).\n

Previous papers related to some features of the package are:

-     J. Castro, A specialized interior-point algorithm for multicommodity network flows, 
      SIAM Journal on Optimization, 10 (2000), 852-877.\n
-     J. Castro, An interior-point approach for primal block-angular problems,
      Computational Optimization and Applications 36 (2007) 195-219.\n 
-     J. Castro, J. Cuesta, Quadratic regularizations in an interior-point method 
      for primal block-angular problems, Mathematical Programming, 130 (2011) 415-445.\n

Some features of the implementation are:

 - it can deal with any problem with primal block-angular structure  \n
 - it deals with quadratic costs of variables \n
 - it considers linear and quadratic costs for slacks of linking constraints \n
 - it considers lower and upper bounds for linking constraints \n
 - rhs terms b_i and b_{k+1} can not be infinity.
 - it can deal with convex separable nonlinear problems (i.e., with diagonal Hessian) \n
 - both Newton and second-order predictor-corrector directions are available \n

The user can also solve the dual direction systems by the Cholesky factorization
of the full system A*Theta*A' (A includes block and linking constraints).
 
Instances are created by the BlockIP() constructors. There are two variants of 
the constructor: one for problems in standard form (0 <= variables <= ub and 
constraints = rhs, and linking constraints contain slacks), 
another for general problems (lb <= variables <= ub and lhs <= constraints  <= rhs). 

See the constructors for information about the input parameters needed to define 
an instance.

Original idea and implementation: Jordi Castro \n
Other programmers: Xavi Jimenez


Jordi Castro, May 2014 \n
Dept. of Statistics and Operations Research \n
Universitat Politecnica de Catalunya

Barcelona, Catalonia
*****************************************************************************/
#ifndef BLOCKIP_H
#define BLOCKIP_H

#include "MatrixBlockIP.h"
#include "StdForm.h"
#include <limits> // for numeric_limits
#include <climits> // for C numeric_limits
#ifdef OPENMP
#include <omp.h>
#endif

using namespace std;

//! Main class for loading and solving problems
class BlockIP {
    
    // data representation
    
 public: // some types needed by users of the class
  enum TYPE_PROBLEM {LINEAR, QUADRATIC, NONLINEAR};
  /*!
    Different options to compute direction dy  according to TYPE_COMP_DY. Note that for options 
    0,2-7, Cholesky is used for blocks, and PCG for linking, they differ in the preconditioner used.
    
    0. CHOL_PWRS_PC: Power Series preconditioner.
    1. FULL_CHOL: Cholesky for full matrix of constraints.
    2. HYBRID_PCG: Power Series + splitting preconditioner (switch according to spectral radius).
    3. CHOL_THETA0_PCG: Theta0 preconditioner for all IPM iterations.
    4. PWRS_THETA0_PCG: Power Series preconditioner + Theta0 preconditioner.
    5. THETA0_PWRS_PCG: Theta0 preconditioner + Power Series preconditioner.
    6. AUT_PWRSTHETA0_PCG: automatic selection of initial preconditioner to be used (either Theta0 or power series),
       afterwards it switches to the other preconditioner. Initial selection based on principal angles.
    7. DYN_PWRSTHETA0_PCG: dynamic selection of wich preconditioner to be used  
       (either Theta0 or power series), every num_it_dyn_PCG number of iterations, based on principal 
       angles and/or spectral radius.   
   */
  enum TYPE_COMP_DY {CHOL_PWRS_PCG, FULL_CHOL, HYBRID_PCG, CHOL_THETA0_PCG, PWRS_THETA0_PCG, THETA0_PWRS_PCG, AUT_PWRSTHETA0_PCG, DYN_PWRSTHETA0_PCG};
  // AUTOMATIC is NEWTON if NONLINEAR problem or PCG solver is used
  enum TYPE_DIRECTION {NEWTON,SECOND_ORDER,AUTOMATIC}; 
  enum TYPE_START_POINT {SIMPLE, QUAD_EQCONS_PROB};
  // QUAD_REG: regularization term f(mu) * 1/2 x'Qx is added
  // PROX_REG: proximal point regularization term f(mu) * 1/2* (x-xk)'Q(x-xk) is added
  enum TYPE_REG {NO_REG, QUAD_REG, PROX_REG};
  enum OUTPUT {NONE, SCREEN, FILE, BOTH};
  /*! two types of computation of principal angles: 
    - principal angle of the linear combination of linking for each block
    - or average of principal angles for all linking of each block
  */
  enum TYPE_COMP_ANGLE {LINEARCOMB,AVERAGE};
  // how to deal with free variables with 0 Theta entries: either splitting them or regularizing its Theta entries
  enum TYPE_DEAL_FREEVARS {SPLIT,REGULARIZE};

 private:

    // constants
    static constexpr double INF= 1.0e128;
    static const int M_PW_PREC= 1;
    static constexpr double SIGMA= 0.1;
    static constexpr double RHO= 0.995;
    static constexpr double OPTIM_GAP= 1.0e-6;
    static constexpr double OPTIM_GAP_SAFEGUARD= 1.0e-5;
    static constexpr double OPTIM_PFEAS= 1.0e-6;
    static constexpr double OPTIM_DFEAS= 1.0e-6;
    static const int OUTPUT_FREQ= 1;
    static const int MAXITER= 200;
    static constexpr double PCG_TOL_LIN= 1.0e-2;
    static constexpr double PCG_TOL_QUAD= 1.0e-3;
    static constexpr double MIN_PCGTOL= 1.0e-8;
    static constexpr double RED_PCGTOL= 0.95;
    static const TYPE_COMP_DY TYPE_COMP_DY_DEFAULT= CHOL_PWRS_PCG;
    static const TYPE_DIRECTION TYPE_DIRECTION_DEFAULT= AUTOMATIC;
    static const TYPE_START_POINT TYPE_START_POINT_DEFAULT= SIMPLE;
    static const TYPE_REG TYPE_REG_DEFAULT= NO_REG;
    static constexpr double FACTOR_REG_DEFAULT= 1.0e-6;
    static const bool DEACTIVATELNK= true;
    static const int MAX_ALLOWED_ZERO_PIVOTS= INT_MAX; //numeric_limits<int>::max();
    static const CHOL_SOLVER CHOL_SOLVER_DEFAULT= SPRSBLKLLT;
    static constexpr double THRESHOLD_ANGLE_DEFAULT= 0.098175; // = pi/32
    static constexpr double THRESHOLD_SPECRAD_DEFAULT= 0.95; // close to 1
    static const int IT_THETA0PWRS_DEFAULT= 3;
    static const TYPE_COMP_ANGLE TYPE_COMP_ANGLE_DEFAULT= LINEARCOMB;
    static const TYPE_DEAL_FREEVARS DEAL_FREEVARS_DEFAULT= REGULARIZE;
    static constexpr double REG_FREEVARS= 1.0e-8;
    static const int NUM_THREADS_DEFAULT= 1;

    // definition of problem, variables provided by the user and other not provided 
    // but required and computed
    int input_problem; // -1 unchecked problem, 0 wrong problem, 1 ok problem
    
//only to test
public:
    int k_blocks; //!< Number of diagonal blocks
    int km; //!< Number of constraints without linking constraints
    int kn; //!< Number of variables without slacks of linking
    int m_cons; //!< Number of constraints including linking constraints
    int n_vars; //!< Number of variables including slacks
    int l_link; //!< Number of linking constraints
    int *ini_n; //!< Begin to blocks 1..k and linking for variables 
    int *ini_m; //!< Begin to blocks 1..k and linking for constraints
    double *cost; //!< Linear cost of variables including slacks
    double *qcost; //!< Quadratic cost of variables including slacks
    double *lb; //!< Lower bounds, including slack bounds
    double *ub; //!< Upper bounds, including slack bounds
    double *lhs; //!< Lower constraint limits, including linking constraints limits
    double *rhs; //!< Upper constraint limits, including linking constraints limits
    
    bool sameN; //!< Define if the same matrix is used for each N block
    bool sameL; //!< Define if the same matrix is used for each L block
    MatrixBlockIP *N; //!< Diagonal blocks
    MatrixBlockIP *L; //!< Linking constraints blocks
    MatrixBlockIP *A; //!< Full matrix (likely made from N and L)
    MatrixBlockIP *D; //!< D = Theta_(k+1) + sum{i..k} L_i*Theta_i*L_i'
    MatrixBlockIP *Theta0; //!< Stores Theta0 preconditioner (Theta_(k+1)) of active linking constraints)
    //! User function to calculate the objective function in a point
    void (*fobj) (int n, double x[], double &fx, double Gx[], double Hx[], void *params);
    void *params; //!< User parameters to perform objective function calculations
    double fx, *Gx, *Hx; //!< Objective function value, Gradient and Hessian in the actual point
    double dobj; //!< dual objective
    
    bool inStdForm; //!< To control if the problem is in standard form

    string *blockNames; //!< Block names of N blocks
    string *varNames; //!< Variable names including slacks
    string *consNames; //!< Constraint names including linking constraints
    
    TYPE_PROBLEM type_objective; //!< Type of objective function, linear, quadratic or non-linear
    StdForm *stdForm; //!< Contains all the information to perform conversions between the original and standard problem 
    double constantFObj; //!< Constant to add in the objective function
    int *origNVars; //!< Number of original variables for each block
    int *origNCons; //!< Number of original constraints for each block
    CHOL_SOLVER chol_solver; //! Cholesky solver to be used (User interface)

private:
//end only to test
    bool initialized; //!< To control if the attributes have been initialized
    
    bool keepStdFormAfterDelete; //!< To keep stdForm when BlockIP will delete
    
    bool deleteMatrices; //!< To delete matrices when created by BlockIP (read_mps)
    
    /*! for matrix factorizations */
    WHO_PERMUTES whoperm;
    
    /*! For the default values see the constant terms in BlockIP.h */
    double inf; //!< Infinity value

    double optim_gap; //!< Optimality gap tolerance
    double optim_pfeas; //!< Primal feasibility tolerance
    double optim_dfeas; //!< Dual feasibility tolerance
    int maxiter; //!< Maximum number of IP iterations

    int output_freq; //!< Output information lines will be printed each output_freq IP iterations
    OUTPUT output; //!< Type of output

    double epsmach; //!< Stores computed machine epsilon
    double init_pcgtol; //!< Initial tolerance for the conjugate gradient (by default PCG_TOL_LIN if linear problem, PCG_TOL_QUAD if quadratic or -convex- nonlinear)
    double pcgtol; //!< Tolerance for conjugate gradient at current iteration
    double min_pcgtol; //!< Minimum tolerance for the conjugate gradient
    double maxit_pcg; //!< Maximum number of pcg iterations (if 0, then the value will be computed by the code)
    double red_pcgtol; //!< Reduction of pcg_tol at each IP iteration
    int m_pw_prec; //!< Number of terms of the power series expansion of (D-C'*B^{-1}*B)^{-1} used as preconditioner
    int totcgit; //!< Overall number of CG iterations
    int cgit; //!< number of CG iterations of this IPM iteration
    int it; //!< IPM iterations
    double est_specrad; //!< Estimation of spectral radius of D^{-1}C'*B^{-1}*B (computed through Ritz values)
    bool show_specrad; //! Whether the estimation of spectral radius has to be shown in the output
    bool comp_specrad; //! Whether the estimation of spectral radius has to be computed
    bool show_princ_angles; //! Whether principal angles between L and N are shown in the output
    double gap_change_to_Chol; //! Change to Cholesky if duality gap below this value
    int allowed_zero_pivots; //! Max number of zero pivots allowed in Cholesky factorization before stopping optimization
    int it_zero_pivots; //! Number of zero pivots of current IP iteration
    bool show_zero_pivots; //! Show number of zero pivots found at each IP iteration
    bool switched_to_Chol; //! Whether we have switched from some other method to full Cholesky for computing directions
    bool PCG_failed; //! True if PCG failed in solving normal equations part of linking constrains
    bool stop_if_PCG_fails; //! Additional stopping criterion: if PCG failed, algorithm stops with current solution without changing to Newton (useful for very large instances)
    int it_Theta0PWRS; //! Decide to change between PWRS and THETA0 every it_Theta0PWRS iterations
    int num_switches_Theta0PWRS; //! Number of switches between PWRS to THETA0 or THETA0 to PWRS
    TYPE_COMP_ANGLE type_comp_angle; //! type of computation of principal angles (linear combination or average)

    TYPE_COMP_DY type_comp_dy; //!< How dy direction is computed; value given by the user
    TYPE_COMP_DY this_type_comp_dy; //! Type of computation of dy direction of this particular IPM iteration
    TYPE_DIRECTION type_direction; //!< Type of direction (Newton, second order...) given by user
    TYPE_DIRECTION this_type_direction; //!< Type of direction (Newton or second order) of this particular IPM iteration
    TYPE_START_POINT type_start_point; //! How starting point will be computed 
    TYPE_REG type_reg; //! Type of regularization performed
    double factor_reg; //! initial value for regularization

    double *x, *y, *z, *w, *s; //!< primal and dual optimization variables
    double *rb, *rc, *rxz, *rsw; //!< residuals of KKT conditions 
    double *dxdz, *dsdw; //!< for rhs of corrector system, from dx,dw,ds and dw of predictor direction
    double normb, normc, normrb, normrc; //!< norms of rhs, costs, rb, and rc
    double alpha_p, alpha_d; //!< step lengths
    double rho; //!< Reduction of the step-length for the primal and dual variables at each IP iteration
    double *dx, *dy, *dz, *dw; //!< Primal and dual optimization directions
    double gap; //!< Duality gap
    double mu, mu0; //!< Centrality parameter; mu0 is value of mu at starting point
    double sigma; //!< Reduction of the centrality parameter at each IP iteration
    double psi; //!< Reduction of dx*dz vector in the rhs of corrector system (in predictor-corrector)
    double *Theta; //!< Theta diagonal matrix
    double *Ax, *Aty, *r, *rhsdy, *rhspcg, *Cvaux, *Cvaux2, *Ctvaux, *auxm, *aux2m; //!< Auxiliary vectors for interior-point algorithm
    double *r_cg, *z_cg, *p_cg; //!< Auxiliary vectors for PCG
    double qreg; //!< Regularization term
    double *valpha, *vbeta; //! Vectors to store information from PCG to later compute Ritz values
    double avg_cos_princ_angle, avg_princ_angle; //! Store information about principal angles (in radians) and its cosine (averaged values for all blocks)
    double threshold_angle; //! threshold principal angle (in radians) to switch between PWRS and THETA0 preconditioners
    double threshold_specrad; //! threshold spectral radius (in (0,1)) to switch between PWRS and THETA0 preconditioners
    bool numfact_NNt_done_iter; //! whether numerical factorization has been performed at each IPM iteration

    bool deactivateLnk; //!< Setting deactivateLnk= false the user may avoid the deactivation of linking
    int numActiveLnk; //!< Number of active lnk
    int numInactiveLnk; //!< Number of inactive lnk
    int numInactiveLnkWithUb; //!< Number of inactive lnk with upper bound
    bool *isActiveLnk; //!< linking i is active if isActiveLnk[i] is true
    int *listActiveLnk; //!< list of active linking, j=listActiveLnk[i], i=1..numActiveLnk, j in {1,..,l_link})
    int *listInactiveLnk; //!< list of inactive linking, j=listInactiveLnk[i], i=1..numInactiveLnk, j in {1,..,l_link})
    
    TYPE_DEAL_FREEVARS deal_freevars; //! how to deal with free vars (splitting or regularizing)
    int numFreeVars; //!< Number of variables marked as free
    int numNonfreeVars; //!< Number of variables not marked as free
    int numFreeVarsRegul; //! Number of subset of free variables with 0 Hessian that need regularization
    bool *isFreeVar; //!< Variable i is marked as free if isFreeVar[i] is true
    int *listFreeVars; //!< List of variables marked as free, j=listFreeVars[i], i=1..numFreeVars, j in {1,..,n_vars})
    int *listNonfreeVars; //!< List of variables not marked as free, j=listNonfreeVars[i], i=1..numNonfreeVars, j in {1,..,n_vars})
    int *listFreeVarsRegul; //!< List of numFreeVarsRegul free variables with 0 Hessian that need regularization (it is a subset of listFreeVars) 
    int numSplitVars; //! Number of pairs of (x+,x-) variables split in standard form, from free variables
    // locSplit_xplus[i] and locSplit_xplus[i] provide location of (x+[i],x-[i]) in vector x, for i=1:numSplitVars
    int * locSplit_xplus; //! locations of variables x+ in standard form, from free variables
    int * locSplit_xminus; //! locations of variables x- in standard form, from free variables
    struct BackupLnk {
      int lnk;
      double rhs;
      double **a;
    };
    int numBackupLnk;
    BackupLnk *backupLnk;
    
    int numWithUb; //!< Number of variables with upper bound (not infinity)
    int numWithoutUb;
    int *listWithoutUb;
    int *listWithUb;

    /*// BlockIPminimize variables
    double *b;
    double *u;
    double *c;
    double *q;*/

    int num_threads; //!< Number of threads set by user
    int num_threads_opt; //!< Number of threads used during optimization (computed from num_threads and max. number of cores)
    double wclocktime; //! wall clock time spent in optimization (computed and updated after each call to minimize() function)

    // member functions

 public: //interface routines
    
    // Constructor
    BlockIP();
    
    // Create an instance with problem in standard form: 0 <= variables <= ub and constraints = rhs
    BlockIP(TYPE_PROBLEM type_objective, double* &cost, double* &qcost, void (*fobj) (int n, double x[], double &fx, double Gx[], double Hx[], void *params), void *params, double* &ub, double* &rhs, int numBlocks, bool sameN, MatrixBlockIP N[], bool sameL, MatrixBlockIP L[]);
    
    // Create an instance with general problem: lb <= variables <= ub and lhs <= constraints  <= rhs
    BlockIP(TYPE_PROBLEM type_objective, double* &cost, double* &qcost, void (*fobj) (int n, double x[], double &fx, double Gx[], double Hx[], void *params), void *params, double* &lb, double* &ub, double* &lhs, double* &rhs, int numBlocks, bool sameN, MatrixBlockIP N[], bool sameL, MatrixBlockIP L[]);
    
    // Destructor
    ~BlockIP();

    // Initializes class attributes
    void initialize();
    // Frees memory of class atrributes
    void free_memory();

    // Create a problem in standard form: 0 <= variables <= ub and constraints = rhs
    void create_problem(TYPE_PROBLEM type_objective, double* &cost, double* &qcost, void (*fobj) (int n, double x[], double &fx, double Gx[], double Hx[], void *params), void *params, double* &ub, double* &rhs, int numBlocks, bool sameN, MatrixBlockIP N[], bool sameL, MatrixBlockIP L[]);
    
    // Create a general problem: lb <= variables <= ub and lhs <= constraints  <= rhs
    void create_problem(TYPE_PROBLEM type_objective, double* &cost, double* &qcost, void (*fobj) (int n, double x[], double &fx, double Gx[], double Hx[], void *params), void *params, double* &lb, double* &ub, double* &lhs, double* &rhs, int numBlocks, bool sameN, MatrixBlockIP N[], bool sameL, MatrixBlockIP L[]);
    
/*    // Load a problem in standard form: 0 <= variables <= ub and constraints = rhs
    void load_problem(double cost[], double qcost[], void (*fobj) (int n, double x[], double &fx, double Gx[], double Hx[], void *params), void *params, double ub[], double rhs[], int numBlocks, bool sameN, MatrixBlockIP N[], bool sameL, MatrixBlockIP L[], bool copy_vectors=true);
    
    // Load a general problem: lb <= variables <= ub and lhs <= constraints  <= rhs
    void load_problem(double cost[], double qcost[], void (*fobj) (int n, double x[], double &fx, double Gx[], double Hx[], void *params), void *params, double lb[], double ub[], double lhs[], double rhs[], int numBlocks, bool sameN, MatrixBlockIP N[], bool sameL, MatrixBlockIP L[], bool copy_vectors=true, bool convertToStd=true);*/
    
    // Convert a problem into standard form
    void convert_to_standard();

    // Check the input problem and compute dimensions
    void check_input_problem_and_compute_dimensions();
    
    // Check if the problems satisfies the conditions when using sameN
    void check_sameN();
    
    // Problem minimization
    int minimize();
    // set to default values all the parameters of the minimization interior-point algorithm
    void set_defaults_minimize();

    /* Retrieve results after optimization */

    // Return number of IP iterations
    int get_it();

    // Return overall number of PCG iterations
    int get_cgit();

    // Return objective function
    double get_fobj();
    
    // Return dual objective function
    double get_dobj();
    
    /* Change some parameters */    
    // Set infinity value to be considered
    void set_inf(double inf= INF);
    
    // Get infinity value to be considered
    double get_inf();
    
    // Set number of terms used as preconditioner of the power series expansion of (D-C'*B^{-1}*B)^{-1}
    void set_m_pw_prec(int m_pw_prec=M_PW_PREC);

    // Get number of terms used as preconditioner of the power series expansion of (D-C'*B^{-1}*B)^{-1}
    int get_m_pw_prec();

    // Set reduction of the centrality parameter at each IP iteration
    void set_sigma(double sigma=SIGMA);
    
    // Get reduction of the centrality parameter at each IP iteration
    double get_sigma();
    
    // Set reduction of the step-length for the primal and dual variables at each IP iteration
    void set_rho(double rho=RHO);
    
    // Get reduction of the step-length for the primal and dual variables at each IP iteration
    double get_rho();
    
    // Set optimality gap tolerance
    void set_optim_gap(double optim_gap=OPTIM_GAP);
    
    // Get optimality gap tolerance
    double get_optim_gap();
    
    // Set primal feasibility tolerance
    void set_optim_pfeas(double optim_pfeas=OPTIM_PFEAS);
    
    // Get primal feasibility tolerance
    double get_optim_pfeas();
    
    // Set dual feasibility tolerance
    void set_optim_dfeas(double optim_dfeas=OPTIM_DFEAS);
    
    // Get dual feasibility tolerance
    double get_optim_dfeas();
    
    // Set output information lines will be printed each output_freq IP iterations
    void set_output_freq(int output_freq=OUTPUT_FREQ);
    
    // Get output information lines will be printed each output_freq IP iterations
    int get_output_freq();
    
    // Set type of output
    void set_output(OUTPUT output=SCREEN);
    
    // Get type of output
    OUTPUT get_output();
    
    // Set maximum number of IP iterations
    void set_maxiter(int maxiter=MAXITER);
    
    // Get maximum number of IP iterations
    int get_maxiter();
    
    // Set initial tolerance for the conjugate gradient (by default PCG_TOL_LIN if linear problem, PCG_TOL_QUAD if quadratic or -convex- nonlinear)
    void set_init_pcgtol(double init_pcgtol);
    
    // Get initial tolerance for the conjugate gradient (by default PCG_TOL_LIN if linear problem, PCG_TOL_QUAD if quadratic or -convex- nonlinear)
    double get_init_pcgtol();
    
    // Set minimum tolerance for the conjugate gradient
    void set_min_pcgtol(double min_pcgtol=MIN_PCGTOL);
    
    // Get minimum tolerance for the conjugate gradient
    double get_min_pcgtol();
    
    // Set maximum number of pcg iterations (if 0, then the value will be computed by the code)
    void set_maxit_pcg(double maxit_pcg=0);
    
    // Get maximum number of pcg iterations (if 0, then the value will be computed by the code)
    double get_maxit_pcg();
    
    // Set reduction of pcg_tol at each IP iteration
    void set_red_pcgtol(double red_pcgtol=RED_PCGTOL);
    
    // Get reduction of pcg_tol at each IP iteration
    double get_red_pcgtol();
    
    // Set show_specrad at each IP iteration
    void set_show_specrad(bool show_specrad=false);
    
    // Get show_specrad 
    bool get_show_specrad();

    // Set show_princ_angles 
    void set_show_princ_angles(bool show_princ_angles=false);
    
    // Get show_princ_angles
    bool get_show_princ_angles();
    
    // Set gap_change_to_Chol (default is 0, never change to Cholesky by this criterion)
    void set_gap_change_to_Chol(double gap_change_to_Chol=0.0);
    
    // Get gap_change_to_Chol
    double get_gap_change_to_Chol();

    // Set allowed_zero_pivots (default is maximum integer, so this stopping criterion is not active)
    void set_allowed_zero_pivots(int allowed_zero_pivots=MAX_ALLOWED_ZERO_PIVOTS);
    
    // Get allowed_zero_pivots
    int get_allowed_zero_pivots();

    // Set show_zero_pivots at each IP iteration
    void set_show_zero_pivots(bool show_zero_pivots=false);
    
    // Get show_zero_pivots 
    bool get_show_zero_pivots();

    // Set stop_if_PCG_fails
    void set_stop_if_PCG_fails(bool stop_if_PCG_fails=false);
    
    // Get stop_if_PCG_fails
    bool get_stop_if_PCG_fails();

     // Set how dy direction is computed
    void set_type_comp_dy(TYPE_COMP_DY type_comp_dy=TYPE_COMP_DY_DEFAULT);
    
    // Get how dy direction is computed
    TYPE_COMP_DY get_type_comp_dy();
    
    // Set which direction is computed
    void set_type_direction(TYPE_DIRECTION type_direction=TYPE_DIRECTION_DEFAULT);
    
    // Get which direction is computed
    TYPE_DIRECTION get_type_direction();
    
     // Set how starting point is computed 
    void set_type_start_point(TYPE_START_POINT type_start_point=TYPE_START_POINT_DEFAULT);
    
    // Get how starting point is computed
    TYPE_START_POINT get_type_start_point();
    
     // Set type of regularization
    void set_type_reg(TYPE_REG type_reg=TYPE_REG_DEFAULT);
    
    // Get type of regularization
    TYPE_REG get_type_reg();
 
     // Set factor of regularization
    void set_factor_reg(double factor_reg=FACTOR_REG_DEFAULT);
    
    // Get factor of regularization
    double get_factor_reg();
 
    // Set flag on deactivation of linking
    void set_deactivateLnk(bool deactivateLnk=DEACTIVATELNK);
    
    // Get flag on deactivation of linking
    bool get_deactivateLnk();

    // Set cholesky solver to be used
    void set_chol_solver(CHOL_SOLVER chol_solver=CHOL_SOLVER_DEFAULT);
    
    // Get Cholesky solver
    CHOL_SOLVER get_chol_solver();

    // Set threshold principal angle to switch between PWRS and THETA0
    void set_threshold_angle(double threshold_angle=THRESHOLD_ANGLE_DEFAULT);
    
    // Get threshold principal angle 
    double get_threshold_angle();
    
    // Set threshold spectral radius to switch between PWRS and THETA0
    void set_threshold_specrad(double threshold_specrad=THRESHOLD_SPECRAD_DEFAULT);
    
    // Get threshold spectral radius 
    double get_threshold_specrad();

    // Set number of iterations between checks to switch between PWRS and THETA0
    void set_it_Theta0PWRS(int it_Theta0PWRS= IT_THETA0PWRS_DEFAULT);
    
    // Get number of iterations between checks to switch between PWRS and THETA0
    int  get_it_Theta0PWRS();

     // Set how principal angle is computed
    void set_type_comp_angle(TYPE_COMP_ANGLE type_comp_angle=TYPE_COMP_ANGLE_DEFAULT);
    
    // Get how principal angle is computed
    TYPE_COMP_ANGLE get_type_comp_angle();
 
    // Set how free variables with 0 Theta entries are dealt with
    void set_deal_freevars(TYPE_DEAL_FREEVARS deal_freevars= DEAL_FREEVARS_DEFAULT);
    
    // Get how free variables with 0 Theta entries are dealt with
    TYPE_DEAL_FREEVARS get_deal_freevars();


    // Get the number of blocks
    int get_k_blocks();
    
    // Get the number of constraints without linking constraints
    int get_km();

    // Get the number of variables without linking constraints
    int get_kn();

    // Get the number of linking constraints
    int get_l_link();

    // Get the number of variables of the problem to be optimized
    int get_n_vars();
    
    // Get the number of variables of the problem to be optimized
    int get_m_cons();

    // Get the number of variables of the original problem
    int get_num_vars();
    
    // Get the number of constraints of the original problem
    int get_num_cons();
    
    // Set who performs the permutation
    void set_whoperm(WHO_PERMUTES whoperm_=CHOLESKY);
    
    // Get who performs the permutation
    WHO_PERMUTES get_whoperm();
    
    // Get primal variables of problem optimized
    const double* get_x();

    // Get dual variables of problem optimized (of equality constraints)
    const double* get_y();

    // Get dual variable of problem optimized (of x>=0) i=1..n_vars
    const double* get_z();

    // Get dual variable of problem optimized (of x<=u) i=1..n_vars
    const double* get_w();

    // Get value of primal variable x(i) of problem optimized i=1..n_vars
    double get_x(int i);

    // Get value of dual variable y(i) of problem optimized (of equality constraints) i=1..m_cons
    double get_y(int i);

    // Get value of dual variable z(i) of problem optimized (of x>=0) i=1..n_vars
    double get_z(int i);

    // Get value of dual variable w(i) of problem optimized (of x<=u) i=1..n_vars
    double get_w(int i);

    // These two needed public for wrapper functions to call PCG
    int min_PCG_Hv(int nn, double *v, double *Hv);
    int min_PCG_Hz_eq_r(int nn, double *zz, double *rr);
    int min_PCG_Theta0z_eq_r(int nn, double *zz, double *rr);

    // Set the constant to add to the objective function
    void set_constant_fobj(double constant);
    
    // Get the constant to add to the objective function
    double get_constant_fobj();
    
    // Set the names of the problem
    void set_names(string *blockNames, string *varNames, string *consNames, bool copy_vectors=true);
    
    // Get the names of the problem
    void get_names(const string* &blockNames, const string* &varNames, const string* &consNames);
    
    // Convert the problem to standard form (if it is not already converted) and write the problem in mps file
    void convert_to_std_and_write_mps(const char *filename);
    
    // Write the problem in mps file
    void write_mps(const char *filename);
    
    // Write the problem in mps file
    void write_mps(const char *filename, TYPE_PROBLEM type_objective, double cost[], double qcost[], double lb[], double ub[], double lhs[], double rhs[], int numBlocks, bool sameN, MatrixBlockIP N[], bool sameL, MatrixBlockIP L[], string *blockNames=NULL, string *varNames=NULL, string *consNames=NULL, double constantFObj=0);

    // Create a problem from a mps file
    void read_mps(const char *filename);
    
    // Write the problem in BlockIP format file
    void write_BlockIP_format(const char *filename);
    
    // Write matrices and vectors of the problem in matlab format
    void write_matlab_format(const char *filename);

    // Create a problem from a BlockIP format file
    void read_BlockIP_format(const char *filename);
    
    // Write all data related to the problem into a file
    void write_problem(const char *filename);
    
    // Get the StdForm related to the problem
    StdForm* get_std_form(bool keepStdFormAfterDelete=false);
  
    // Create names for blocks, variables and constraints if does not exist
    void create_names();
    
    // Get the name of the variables including the block name
    string* get_full_var_names();
    
    // Get the name of the constraints including the block name
    string* get_full_cons_names();

    // Set number of threads to be used
    void set_num_threads(int num_threads);

    // Get number of threads to be used
    int get_num_threads();
    
    // Get max number of threads/cores available (if not changed by a call to omp_set_num_threads())
    int get_max_num_threads();

    // Get wall clock time (computed by minimize(), only when OpenMP is used)
    double get_wclock_time();

 private:
    void min_manage_threads();
    void min_initializations();
    void min_preprocess();
    void min_normb_normc();
    void min_Ax(double *Ax, const double *x);
    void min_Aty(double *Aty, const double *y);
    void min_compute_s();
    void min_compute_mu();
    void min_compute_sigma_psi();
    void min_update_inactivelnk();
    void min_KKT_residuals();
    void min_starting_point();
    void min_objective_functions_nonlin();
    void min_objective_functions_linquad();
    void eval_fobj(double x[], double &fx, double Gx[], double Hx[]);
    void min_write_problem_information();
    void min_write_current_point_info();
    void min_steplengths(double &talpha_p, double &talpha_d, double tdx[], double tdz[], double tdw[]);
    bool min_optimal();
    void min_compute_Theta();
    void min_compute_Theta0();
    void min_compute_direction();
    void min_Newton_direction();
    void min_second_order_predictor_corrector_direction();
    void min_update_newpoint();
    void min_test_newpoint();
    void min_backup_inactive(int lnk);
    void min_restore_inactive(int lnk);
    void min_compute_dy_cholpcg(double *dy, double *rhsdy, bool only_solve=false);
    void min_compute_dy_fullchol(double *dy, double *rhsdy, bool only_solve=false);
    void min_free_memory();
    void min_check_Newton_direction_is_correct();
    void min_check_predictor_direction_is_correct();
    void min_check_predictor_corrector_direction_is_correct();
    void min_solve_NThetaNt(double *v);
    void min_Ctv(double *v, double *Ctv);
    void min_Cv(double *v, double *Cv);
    void min_debug_write_variables();
    void min_debug_variables_of_inactivelnk();
    void min_debug_variables_without_ub();
    void min_update_qreg();
    void min_compute_princ_angles(bool isTheta1, bool only_solve=false, TYPE_COMP_ANGLE type_angle= TYPE_COMP_ANGLE_DEFAULT);
    TYPE_COMP_DY switch_preconditioners();
};

inline void BlockIP::set_inf(double inf)
//! Set threshold value to be used as infinity (x>inf -> x is inf)
{
    this->inf= inf;
}

inline double BlockIP::get_inf()
//! Get infinity value to be considered
{
  return inf;
}

inline void BlockIP::set_m_pw_prec(int m_pw_prec)
//! Set number of terms used as preconditioner of the power series expansion of (D-C'*B^{-1}*B)^{-1}
{
  this->m_pw_prec= m_pw_prec;
}

inline  int BlockIP::get_m_pw_prec()
//! Get number of terms used as preconditioner of the power series expansion of (D-C'*B^{-1}*B)^{-1}
{
  return m_pw_prec;
}

inline void BlockIP::set_sigma(double sigma)
//! Set reduction of the centrality parameter at each IP iteration
{
  this->sigma= sigma;
}

inline double BlockIP::get_sigma()
//! Get reduction of the centrality parameter at each IP iteration
{
  return sigma;
}

inline void BlockIP::set_rho(double rho)
//! Set reduction of the step-length for the primal and dual variables at each IP iteration
{
  this->rho= rho;
}

inline double BlockIP::get_rho()
//! Get reduction of the step-length for the primal and dual variables at each IP iteration
{
  return rho;
}

inline void BlockIP::set_optim_gap(double optim_gap)
//! Set optimality gap tolerance
{
  this->optim_gap= optim_gap;
}

inline double BlockIP::get_optim_gap()
//! Get optimality gap tolerance
{
  return optim_gap;
}

inline void BlockIP::set_optim_pfeas(double optim_pfeas)
//! Set primal feasibility tolerance
{
  this->optim_pfeas= optim_pfeas;
}

inline double BlockIP::get_optim_pfeas()
//! Get primal feasibility tolerance
{
  return optim_pfeas;
}

inline void BlockIP::set_optim_dfeas(double optim_dfeas)
//! Set dual feasibility tolerance
{
  this->optim_dfeas= optim_dfeas;
}

inline double BlockIP::get_optim_dfeas()
//! Get dual feasibility tolerance
{
  return optim_dfeas;
}

inline void BlockIP::set_output_freq(int output_freq)
//! Set output information lines will be printed each output_freq IP iterations
{
  this->output_freq= output_freq;
}

inline int BlockIP::get_output_freq()
//! Get output information lines will be printed each output_freq IP iterations
{
  return output_freq;
}

inline void BlockIP::set_output(OUTPUT output)
//! Set type of output
{
  this->output= output;
}

inline BlockIP::OUTPUT BlockIP::get_output()
//! Get type of output
{
  return output;
}

inline void BlockIP::set_maxiter(int maxiter)
//! Set maximum number of IP iterations
{
  this->maxiter= maxiter;
}

inline int BlockIP::get_maxiter()
//! Get maximum number of IP iterations
{
  return maxiter;
}

inline void BlockIP::set_init_pcgtol(double init_pcgtol)
//! Set initial tolerance for the conjugate gradient (by default PCG_TOL_LIN if linear problem, PCG_TOL_QUAD if quadratic or -convex- nonlinear)
{
  this->init_pcgtol= init_pcgtol;
}

inline double BlockIP::get_init_pcgtol()
//! Get initial tolerance for the conjugate gradient (by default PCG_TOL_LIN if linear problem, PCG_TOL_QUAD if quadratic or -convex- nonlinear)
{
  return init_pcgtol;
}

inline void BlockIP::set_min_pcgtol(double min_pcgtol)
//! Set minimum tolerance for the conjugate gradient
{
  this->min_pcgtol= min_pcgtol;
}

inline double BlockIP::get_min_pcgtol()
//! Get minimum tolerance for the conjugate gradient
{
  return min_pcgtol;
}

inline void BlockIP::set_maxit_pcg(double maxit_pcg)
//! Set maximum number of pcg iterations (if 0, then the value will be computed by the code)
{
  this->maxit_pcg= maxit_pcg;
}

inline double BlockIP::get_maxit_pcg()
//! Get maximum number of pcg iterations (if 0, then the value will be computed by the code)
{
  return maxit_pcg;
}

inline void BlockIP::set_red_pcgtol(double red_pcgtol)
//! Set reduction of pcg_tol at each IP iteration
{
  this->red_pcgtol= red_pcgtol;
}

inline double BlockIP::get_red_pcgtol()
//! Get reduction of pcg_tol at each IP iteration
{
  return red_pcgtol;
}

inline void BlockIP::set_show_specrad(bool show_specrad)
//! Set show_specrad at each IP iteration
{
  this->show_specrad= show_specrad;
}

inline bool BlockIP::get_show_specrad()
//! Get show_specrad
{
  return show_specrad;
}

inline void BlockIP::set_show_princ_angles(bool show_princ_angles)
//! Set show_princ_angles
{
  this->show_princ_angles= show_princ_angles;
}

inline bool BlockIP::get_show_princ_angles()
//! Get show_princ_angles
{
  return show_princ_angles;
}

inline void BlockIP::set_gap_change_to_Chol(double gap_change_to_Chol)
//! Set gap_change_to_Chol (default is 0, never change by this criterion)
{
  this->gap_change_to_Chol= gap_change_to_Chol;
}    

inline double BlockIP::get_gap_change_to_Chol()
//! Get gap_change_to_Chol
{
  return gap_change_to_Chol;
}

inline void BlockIP::set_allowed_zero_pivots(int allowed_zero_pivots)
//! Set allowed_zero_pivots: if number of Cholesky zero pivots greater than this allowed number, then optimization stopped
{
  this->allowed_zero_pivots= allowed_zero_pivots;
}    

inline int BlockIP::get_allowed_zero_pivots()
//! Get allowed_zero_pivots
{
  return allowed_zero_pivots;
}

inline void BlockIP::set_show_zero_pivots(bool show_zero_pivots)
//! Set show_zero_pivots at each IP iteration
{
  this->show_zero_pivots= show_zero_pivots;
}

inline bool BlockIP::get_show_zero_pivots()
//! Get show_zero_pivots
{
  return show_zero_pivots;
}

inline void BlockIP::set_stop_if_PCG_fails(bool stop_if_PCG_fails)
//! Set stop_if_PCG_fails (if true, then stops IPM algorithm if PCG fails with PWRS preconditioner)
{
  this->stop_if_PCG_fails= stop_if_PCG_fails;
}

inline bool BlockIP::get_stop_if_PCG_fails()
//! Get stop_if_PCG_fails
{
  return stop_if_PCG_fails;
}

inline void BlockIP::set_type_comp_dy(TYPE_COMP_DY type_comp_dy)
//! Set how dy direction is computed
{
  this->type_comp_dy= type_comp_dy;
}

inline BlockIP::TYPE_COMP_DY BlockIP::get_type_comp_dy()
//! Get how dy direction is computed
{
  return type_comp_dy;
}

inline void BlockIP::set_type_direction(TYPE_DIRECTION type_direction)
//! Set which direction is computed
{
  this->type_direction= type_direction;
}
 
inline BlockIP::TYPE_DIRECTION BlockIP::get_type_direction()
//! Get which direction is computed
{
  return type_direction;
}    

inline void BlockIP::set_type_start_point(TYPE_START_POINT type_start_point)
//! Set how starting point is computed
{
  this->type_start_point= type_start_point;
}

inline BlockIP::TYPE_START_POINT BlockIP::get_type_start_point()
//! Get how starting point is computed
{
  return type_start_point;
}

inline void BlockIP::set_type_reg(TYPE_REG type_reg)
//! Set type of regularization
{
  this->type_reg= type_reg;
}

inline BlockIP::TYPE_REG BlockIP::get_type_reg()
//! Get type of regularization
{
  return type_reg;
}

inline void BlockIP::set_factor_reg(double factor_reg)
//! Set factor of regularization
{
  this->factor_reg= factor_reg;
}

inline double BlockIP::get_factor_reg()
//! Get factor of regularization
{
  return factor_reg;
}

inline void BlockIP::set_chol_solver(CHOL_SOLVER chol_solver)
//! Set Cholesky solver to be used
{
  this->chol_solver= chol_solver;
}

inline CHOL_SOLVER BlockIP::get_chol_solver()
//! Get the Cholesky solver
{
	return chol_solver;
}

inline void BlockIP::set_deactivateLnk(bool deactivateLnk)
//! Set flag on deactivation of linking
{
  this->deactivateLnk= deactivateLnk;
}
    
inline bool BlockIP::get_deactivateLnk()
//! Get flag on deactivation of linking
{
  return deactivateLnk;
}

inline void BlockIP::set_threshold_angle(double threshold_angle)
//! Set threshold principal angle to switch between PWRS and THETA0
{
  this->threshold_angle= threshold_angle;
}

inline double BlockIP::get_threshold_angle()
//! Get threshold principal angle 
{
  return threshold_angle;
}

inline void BlockIP::set_threshold_specrad(double threshold_specrad)
//! Set threshold spectral radius to switch between PWRS and THETA0
{
  this->threshold_specrad= threshold_specrad;
}

inline double BlockIP::get_threshold_specrad()
//! Get threshold spectral radius
{
  return threshold_specrad;
}

inline void BlockIP::set_it_Theta0PWRS(int it_Theta0PWRS)
//! Set number of iterations between checks to switch between PWRS and THETA0
{
  this->it_Theta0PWRS= it_Theta0PWRS;
}

inline int  BlockIP::get_it_Theta0PWRS()
//! Get number of iterations between checks to switch between PWRS and THETA0
{
  return it_Theta0PWRS;
}

inline void BlockIP::set_type_comp_angle(TYPE_COMP_ANGLE type_comp_angle)
//! Set how principal angle is computed
{
  this->type_comp_angle= type_comp_angle;
}

inline BlockIP::TYPE_COMP_ANGLE BlockIP::get_type_comp_angle()
//! Get how principal angle is computed
{
  return type_comp_angle;
}

inline void BlockIP::set_deal_freevars(TYPE_DEAL_FREEVARS deal_freevars)
//! Set how free variables with 0 Theta entries are dealt with
{
  this->deal_freevars= deal_freevars;
}

inline BlockIP::TYPE_DEAL_FREEVARS BlockIP::get_deal_freevars()
//! Get how free variables with 0 Theta entries are dealt with
{
  return deal_freevars;
}

inline int BlockIP::get_k_blocks()
//! Get the number of blocks
/*!
  @return field k_blocks
*/
{
  return k_blocks;
}

inline int BlockIP::get_km()
//! Get the number of constraints without linking constraints
/*!
  @return field km
*/
{
  return km;
}

inline int BlockIP::get_kn()
//! Get the number of variables without linking constraints
/*!
  @return field kn
*/
{
  return kn;
}

inline int BlockIP::get_l_link()
//! Get the number of linking constraints
/*!
  @return field l_link
*/
{
  return l_link;
}

inline int BlockIP::get_n_vars()
//! Get the number of variables of the problem to be optimized
/*!
  @return field n_vars 
*/
{
  return n_vars;
}

inline int BlockIP::get_m_cons()
//! Get the number of constraints of the problem to be optimized
/*!
  @return field m_cons
*/
{
  return m_cons;
}

inline int BlockIP::get_num_vars()
//! Get the number of variables of the original problem
/*!
  @return Number of variables of the original problem
*/
{
  if (stdForm)
    return stdForm->numOrigVars;
  else
    return n_vars;
}

inline int BlockIP::get_num_cons()
//! Get the number of constraints of the original problem
/*!
  @return Number of constraints of the original problem
*/
{
  if (stdForm)
    return m_cons + stdForm->deletedRows->size();
  else
    return m_cons;
}

inline void BlockIP::set_whoperm(WHO_PERMUTES whoperm_)
//! Set whopermutes constraints-related information: the Cholesky factorization of the user of it
{
    whoperm= whoperm_;
}

inline WHO_PERMUTES BlockIP::get_whoperm()
//! Get whopermutes constraints-related information: the Cholesky factorization of the user of it
{
  return whoperm;
}

inline const double* BlockIP::get_x()
//! Get primal variables of problem optimized
{
  if (stdForm)
    return stdForm->xOrig;
  else
    return x;
}

inline const double* BlockIP::get_y()
//! Get dual variables of problem optimized (of equality constraints)
{
  if (stdForm)
    return stdForm->yOrig;
  else
    return y;
}

inline const double* BlockIP::get_z()
//! Get dual variable of problem optimized (of x>=0) i=1..n_vars
{
  if (stdForm)
    return stdForm->zOrig;
  else
    return z;
}

inline const double* BlockIP::get_w()
//! Get dual variable of problem optimized (of x<=u) i=1..n_vars
{
  if (stdForm)
    return stdForm->wOrig;
  else
    return w;
}

inline double BlockIP::get_x(int i)
//! Get value of primal variable x(i) of problem optimized i=1..n_vars
{
  if (stdForm)
    return stdForm->xOrig[i];
  else
    return x[i];
}

inline double BlockIP::get_y(int i)
//! Get value of dual variable y(i) of problem optimized (of equality constraints) i=1..m_cons
{
  if (stdForm)
    return stdForm->yOrig[i];
  else
    return y[i];
}

inline double BlockIP::get_z(int i)
//! Get value of dual variable z(i) (of x>=0) of problem optimized i=1..n_vars
{
  if (stdForm)
    return stdForm->zOrig[i];
  else
    return z[i];
}

inline double BlockIP::get_w(int i)
//! Get value of dual variable w(i) (of x<=u) of problem optimized i=1..n_vars
{
  if (stdForm)
    return stdForm->wOrig[i];
  else
    return w[i];
}



inline int BlockIP::get_it()
//! Return number of IP iterations of minimization
{
    return(it);
}

inline int BlockIP::get_cgit()
//! Return number of overall PCG iterations
{
    return(totcgit);
}

inline double BlockIP::get_fobj()
//! Return objective function
{
  if (stdForm)
    return fx + stdForm->constantFObj;
  else
    return(fx);
}

inline double BlockIP::get_dobj()
//! Return dual objective function
{
  if (stdForm)
    return dobj + stdForm->constantFObj;
  else
    return(dobj);
}

inline StdForm* BlockIP::get_std_form(bool keepStdFormAfterDelete)
//! Get the StdForm related to the problem
/*!
  @param keepStdFormAfterDelete Set if the StdForm related to the
         problem will be deleted or not when BlockIP will delete
  @return The StdForm related to the problem, if the problem was
          created in standard form, return NULL.
*/
{
  this->keepStdFormAfterDelete = keepStdFormAfterDelete;
  return stdForm;
}

inline void BlockIP::set_constant_fobj(double constant)
//! Set the constant to add to the objective function
{
  constantFObj = constant;
}

inline double BlockIP::get_constant_fobj()
//! Get the constant to add to the objective function
{
  return constantFObj;
}  

inline void BlockIP::eval_fobj(double x[], double &fx, double Gx[], double Hx[])
//! Evaluates objective function, with and without conversion to standard form
{
  if (stdForm)
    stdForm->fobj_stdform(n_vars, x, fx, Gx, Hx, params);
  else
    fobj(n_vars, x, fx, Gx, Hx, params);
}
    
inline void BlockIP::set_num_threads(int num_threads)
//! Set number of threads to be used
{
  this->num_threads= num_threads;
}

inline int BlockIP::get_num_threads()
//! Get number of threads to be used
{
  return num_threads;
}  

inline int BlockIP::get_max_num_threads()
//! It calls OpenMP omp_get_max_threads() if compiled with OpenMP, otherwise 1
//! If no call to omp_set_num_threads() has been performed, it return max CPU cores;
//! otherwise it returns the max number of threads set by a previous call to omp_set_num_threads()
{
#ifdef OPENMP
  return omp_get_max_threads();
#else
  return 1; // sequential execution
#endif
}  
#endif //BLOCKIP_H

inline double  BlockIP::get_wclock_time()
//! Get wall clock time (computed by minimize(), only when OpenMP is used)
{
  return wclocktime;
}
