/*!
  @file MatrixBlockIP.h 
  @brief Definition of MatrixBlockIP
*/

#ifndef MATRIXBLOCKIP_H
#define MATRIXBLOCKIP_H

#include "enums.h"
#include "SparseChol.h"
#include <vector>
#include <fstream>

using namespace std;

//! Class for manipulating matrices, and interfacing SparseChol
class MatrixBlockIP {
 public:
  
  /* 
     For every type of matrix we need routines to:
     - multiply matrix by vector, and matrix transposed by vector
     - solve systems with matrix (either through Cholesky or specialized routines)
     - routine to transform matrix to a GENERAL matrix (to get full matrix of constraints)
  */

  /*! Common in all types matrix */
  TYPE_MATRIX type; //!< Type of matrix
  int m; //!< Number of rows
  int n; //!< Number of columns
  bool ija; //!< True if irowa, icola and a are stored
  bool rowwise; //!< True if inirowa, icola and a are stored
  bool columnwise; //!< True if the matrix is stored in column-wise packed format, incompatible with rowwise
  
  /*! Common packed format attributes */
  int nz; //!< Number of non-zero elements
  double *a; //!< Value of non-zero elements
  
  /*! General row-wise packed format attributes */
  int *inirowa; //!< Index to the first element of each row
  int *icola; //!< Column position for each element
  
  /*! General column-wise packed format attributes */
  int *inicola; //!< Index to the first element of each column
  int *irowa; //!< Row position for each element
  
  /*! Network format attributes */
  TYPE_ORIENTATION type_orientation; //!< Type of orientation, by default, oriented
  int num_arcs; //!< Number of arcs
  int num_nodes; //!< Number of nodes
  int *src; //!< Source of each arc
  int *dst; //!< Destination of each arc
  
  /*! Diagonal format attributes */
  double *d1; //!< d1 for DIAGONAL and 1st submatrix of DIAG_DIAG
  double *d2; //!< d2 for 2nd submatrix of DIAG_DIAG

  /*! For sparse Cholesky */
  SparseChol chol; //!< Sparse Cholesky class
  
  /*! For splitting-type linking constraints matrices, see create_splitting_matrix() for a description of this matrix type */
  int k_split; //!< Number of blocks in splitting matrix (it should be equal to k_blocks variables of class BlockIP)
  int ini_split; //!< Index to first variable in range of splitted variables, same for all blocks 
  int end_split; //!< Index to last variable in range of splitted variables, same for all blocks
  int nnz_cols_split; //! Number of nonzero columns in total splitting matrix: nnz_cols_split= k_split*n_split
  int n_split; //!< Number of splitted variables: n_split= end_split-ini_split+1
  int i_k_general_splitting; //! Number of splitting block transformed to general by a previous call to native_to_general(), from 0 to k_split-1

  /* For symmetric tridiagonal shifted matrix, which appear when computing D=L_{k+1}+sum{i=1..k} L_i*L_i'  and L_i are splliting matrices */
  int t_shift;
  double *tridiag_d; // diagonal values, vector of size m
  double *tridiag_s; // out of diagonal (either upper or lower) values, vector of size m-t_shift;

  /* For dense_diag_diag matrices [A D1 D2] */
  int nd; // number of columns of dense submatrix
  int nzd; // number of nonzeros in dense part
  double *ad; // stores dense matrix A
  bool d1exist; // whether D1 part exist
  bool d2exist; // whether D1 part exist; if d1exist is false, then d2exist assumed to be false too
  double cd1; // constant value of D1
  double cd2; // constant value of D2
  double *d1d; // if d1exist and cd1==0 then stores D1
  double *d2d; // if d2exist and cd2==0 then stores D2


  /*! For D Matrix */
  int sizeL;
  int *inirowLLt;
  // The following vectors are allocated with ALLOC and freed with FREE
  int *icolLLt;
  int *blockD;
  int *inivalD;
  int *icolD;
  double *valD;

  // /*! boolean to check whether symbolic factorization already made */ -> This has to be checked in SparsChol
  // bool made_symbfct_MMt; // for symbolic factorization of M*Theta*M'
  // bool made_symbfct_M;   // for symbolic factorization of M

  /*! boolean to check whether analyze_D already made, needed to compute_D() */
  bool made_analyze_D; 

  /*! number of replications of this matrix; this is needed to store 
    space for num_blocks numerical factorization of M*Theta[i]*M', for different 
    Theta[i], i=0,...,num_blocks-1
    This value is set by the minimization algorithm, which needs the factorizations.
   */
  int num_blocks; 

  // Constructor
  MatrixBlockIP(int blocks=1);
  // Copy constructor
  MatrixBlockIP(MatrixBlockIP *mbip);
  // Destructor
  ~MatrixBlockIP();
  
  // Copy
  void copy(MatrixBlockIP *mbip);
  
  // Comes back to the initial state
  void reset();
  
  // Comes back to the state after create the matrix
  void restore();
  
  /* Native methods to create different matrix types */
  // Native method to create a general row-wise packed matrix
  void create_general_matrix_row_wise(int m, int n, int nz, int* &inirowa, int* &icola, double* &a);
  // Native method to create a general column-wise packed matrix
  void create_general_matrix_column_wise(int m, int n, int nz, int* &inicola, int* &irowa, double* &a);
  // Native method to create a network matrix
  void create_network_matrix(int num_arcs, int num_nodes, int* &src, int* &dst, bool oriented= true);
  // Native method to create a identity matrix
  void create_identity_matrix(int dim);
  // Native method to create a identity-identity matrix
  void create_idty_idty_matrix(int nrows);
  // Native method to create a diagonal matrix
  void create_diagonal_matrix(int dim, double* &d);
  // Native method to create a diagonal-diagonal matrix
  void create_diag_diag_matrix(int nrows, double* &d1, double* &d2);
  // Native method to create a splitting matrix
  void create_splitting_matrix(int k_split, int ini_split, int end_split);
  // Native method to create a dense_diag_diag matrix
  void create_dense_diag_diag_matrix(int m, int nd, double* &a, bool d1exist, bool d2exist, double*& d1, double*& d2, double cd1= 0.0, double cd2= 0.0);

/*  
  // Native method to load a general row-wise packed matrix
  void load_general_matrix_row_wise(int m, int n, int nz, int inirowa[], int icola[], double a[], bool ordered=false, bool copy_vectors=true);
  // Native method to load a general column-wise packed matrix
  void load_general_matrix_column_wise(int m, int n, int nz, int inicola[], int irowa[], double a[], bool ordered=false, bool copy_vectors=true);
  // Native method to load a network matrix
  void load_network_matrix(int num_arcs, int num_nodes, int src[], int dst[], bool oriented= true, bool copy_vectors=true);
  // Native method to load a identity matrix
  void load_identity_matrix(int dim);
  // Native method to load a identity-identity matrix
  void load_idty_idty_matrix(int dim);
  // Native method to load a diagonal matrix
  void load_diagonal_matrix(int dim, double d[], bool copy_vectors=true);
  // Native method to load a diagonal-diagonal matrix
  void load_diag_diag_matrix(int dim, double d1[], double d2[], bool copy_vectors=true);
*/
    
  /* Alternative methods to load/build matrices */
  // Creates a general matrix in format ija
  void create_general_matrix_format_ija(int m, int n, int nz, int* &row_index, int* &col_index, double* &values);
/*
  // Loads a general matrix in format ija
  void load_general_matrix_format_ija(int m, int n, int nz, int row_index[]= NULL, int col_index[]=NULL, double values[]= NULL, bool copy_vectors=true);  
*/
  // Builds a packed rowwise format structured matrix based on diagonal blocks and linking constraints blocks
  void compute_full_matrix(int numBlocks, bool sameN, MatrixBlockIP N[], bool sameL, MatrixBlockIP L[], int *ini_n ,int numActiveLnk=0, int *listActiveLnk=NULL);
  // Create internal structure arrays to compute D = Theta_(numBlocks+1) + sum{i..numBlocks} L_i*Theta_i*L_i'
  void analyze_D(int numBlocks, bool sameL, MatrixBlockIP L[]);
  void analyze_D_diagonal(int numBlocks, bool sameL, MatrixBlockIP L[]);
  void analyze_D_gen_sym_uptr(int numBlocks, bool sameL, MatrixBlockIP L[]);
  void analyze_D_tridiag_shifted(int numBlocks, bool sameL, MatrixBlockIP L[]);
  // Compute D = Theta_(numBlocks+1) + sum{i..numBlocks} L_i*Theta_i*L_i'
  void compute_D(int numBlocks, bool sameL, MatrixBlockIP L[], bool isActive[], int iniTheta[], double Theta[]);
  void compute_D_diagonal(int numBlocks, bool sameL, MatrixBlockIP L[], bool isActive[], int iniTheta[], double Theta[]);
  void compute_D_gen_sym_uptr(int numBlocks, bool sameL, MatrixBlockIP L[], bool isActive[], int iniTheta[], double Theta[]);
  void compute_D_tridiag_shifted(int numBlocks, bool sameL, MatrixBlockIP L[], bool isActive[], int iniTheta[], double Theta[]);
    
  /* Matrix transformations */
  // Calculates and stores the matrix in packed rowwise format
  void native_to_general(bool delete_native_format=false, int i_k= -1, int n_k= -1);
  // Calculates and stores the matrix in packed rowwise format
  void ija_to_rowwise();
  // Calculates and stores the network matrix (either oriented or nonoriented) in packed rowwise format
  void network_to_general();
  // Calculates and stores the I matrix in packed rowwise format
  void identity_to_general();
  // Calculates and stores the diagonal D matrix in packed rowwise format
  void diagonal_to_general();
  // Calculates and stores the matrix [I I] in packed rowwise format
  void idty_idty_to_general();
  // Calculates and stores the matrix [D1 D2] in packed rowwise format
  void diag_diag_to_general();
  // Calculates and stores the i_k-th block of an splitting matrix in packed rowwise format
  void splitting_to_general(int i_k, int n_k);
  // Calculates and stores the dense-diag-diag matrix [D D1 D2] in packed rowwise format
  void dense_diag_diag_to_general();
  
  // Calculates and stores the network matrix in ija format.
  void network_to_ija_format();
  // Calculates and stores the identity I matrix in ija format.
  void identity_to_ija_format();
  // Calculates and stores the diagonal D1 matrix in ija format.
  void diagonal_to_ija_format();
  // Calculates and stores the matrix  [I I] in ija format
  void idty_idty_to_ija_format();
  // Calculates and stores the matrix  [D1 D2] in ija format
  void diag_diag_to_ija_format();
  // Calculates and stores the matrix  [D D1 D2] in ija format
  void dense_diag_diag_to_ija_format();
  // Order the matrix when has been created in row-wise or column-wise format
  void order_matrix();
  
  /* Matrix operations */
  // Matrix-vector product vout=M*vin (vout(m),vin(n))
  void mul_Mv(double vout[], const double vin[], int i_k= 0);
  // MatrixTranspose-vector product vout=M(t)*vin (vout(n),vin(m)
  void mul_Mtv(double vout[], const double vin[], int i_k= 0, int n_vout= 0);
  // Add matrix-vector product vout += M*vin (vout(m),vin(n))
  void add_mul_Mv(double vout[], const double vin[], int i_k= 0);
  // Add matrixTranspose-vector product vout += M(t)*vin (vout(n),vin(m)
  void add_mul_Mtv(double vout[], const double vin[], int i_k= 0, int n_vout= 0);
  // Given a column determines if exists an element for each row of the matrix, in case of exist returns the value.
  void exist_var_in_row(int column, bool appear[], double coefs[]);
  // Adds a new column into the matrix
  void add_new_column(int size, int irows[], double values[]);
  // Adds new columns into the matrix
  void add_new_columns(int num_columns, int size[], int *irows[], double *values[]);
  // Changes the sign of some columns
  void change_columns_sign(int size, int columns[]);
  // Changes the sign of some rows
  void change_rows_sign(int size, int rows[]);
  // Deletes some rows
  void delete_rows(int size, int rows[]);
  // Gets a column of the matrix
  int get_column(int column, int* &irows, double* &values, bool invert_sign= false, int i_k=0);
  // Gets a row of the matrix in unpacked form, multiplying components by scaling Theta and returning norm^2 of scaled row
  int get_unpacked_row_Theta_norm(int row_num,  double row_values[], const double Theta[], double &norm, int nn= 0, int i_k= 0);
  // Gets a row of the matrix in unpacked form
  int get_unpacked_row(int row_num,  double row_values[]);
  // Adds a row of the matrix in unpacked form to a given vector
  int add_unpacked_row(int row_num,  double row_values[]);
  // Compute the sum of rows (only for some special  matrices, like SPLITTING)
  int compute_sum_of_rows(double *row_values, int nn, int i_k= 0);

  // SparseChol interface functions
  // next three for systems with M*Theta*M'
  void symbolic_fact_MMt(CHOL_SOLVER chslv, int *prov_pfa= NULL, int *prov_ipfa= NULL, NUMBERING prov_pfa_numbering= NOT_COMPUTED);
  void numeric_fact_MMt(CHOL_SOLVER chslv, double *Theta, int i_k=0 );
  void numeric_solve_MMt(double *rhs, int i_k= 0, WHO_PERMUTES whoperm= CHOLESKY);
  // next three for systems with M (M positive semidefinite matrix)
  void symbolic_fact_M(CHOL_SOLVER chslv, int *prov_pfa= NULL, int *prov_ipfa= NULL, NUMBERING prov_pfa_numbering= NOT_COMPUTED);
  void numeric_fact_M(CHOL_SOLVER chslv);
  void numeric_solve_M(double *rhs, WHO_PERMUTES whoperm= CHOLESKY);

  // access to data in class
  int get_pfa(int i); // returns SparseChol pfa[i] with no check
  int get_ipfa(int i); // returns SparseChol ipfa[i] with no check
  int get_maxlnz(); // returns SparseChol maxlnz with no check
  int get_maxfillin(); // returns SparseChol maxfillin with no check
  int get_njka(); // returns SparseChol njka with no check
  int get_num_zero_pivots(); // returns SparseChol num_zero_pivots with no check
  int get_num_semidef_matrix(); // returns SparseChol num_semidef_matrix with no check
  
  // only to test
  void print_matrix(bool ija= true, bool start_one= true);
  void print_matrix(ofstream &outfile, bool print_ija= true, bool start_one= true);
  // Print some positions and name of a vector
  void print_vector(double *v, int size, string name);

  // Pre: icola is ordered
  // Convert a matrix in column-wise format to row-wise format.
  void column_wise_to_row_wise_format();
  
  // Pre: irowa is ordered
  // Convert a matrix in row-wise format to column-wise format.
  void row_wise_to_column_wise_format();
    
  private:
  
  // free all the memory allocated by the application - not the user
  void free_memory();

  void mul_Mv_row_wise(double vout[], const double vin[]);
  void mul_Mtv_row_wise(double vout[], const double vin[]);
  void add_mul_Mv_row_wise(double vout[], const double vin[]);
  void add_mul_Mtv_row_wise(double vout[], const double vin[]);
  
  void mul_Mv_column_wise(double vout[], const double vin[]);
  void mul_Mtv_column_wise(double vout[], const double vin[]);
  void add_mul_Mv_column_wise(double vout[], const double vin[]);
  void add_mul_Mtv_column_wise(double vout[], const double vin[]);

  void mul_Mv_network(double vout[], const double vin[]);
  void mul_Mtv_network(double vout[], const double vin[]);
  void add_mul_Mv_network(double vout[], const double vin[]);
  void add_mul_Mtv_network(double vout[], const double vin[]);

  void mul_Mv_identity(double vout[], const double vin[]);
  void add_mul_Mv_identity(double vout[], const double vin[]);

  void mul_Mv_diagonal(double vout[], const double vin[]);
  void add_mul_Mv_diagonal(double vout[], const double vin[]);

  void mul_Mv_idty_idty(double vout[], const double vin[]);
  void mul_Mtv_idty_idty(double vout[], const double vin[]);
  void add_mul_Mv_idty_idty(double vout[], const double vin[]);
  void add_mul_Mtv_idty_idty(double vout[], const double vin[]);

  void mul_Mv_diag_diag(double vout[], const double vin[]);
  void mul_Mtv_diag_diag(double vout[], const double vin[]);
  void add_mul_Mv_diag_diag(double vout[], const double vin[]);
  void add_mul_Mtv_diag_diag(double vout[], const double vin[]);
  
  void mul_Mv_gen_sym_uptr(double vout[], const double vin[]); // M=M', then mul_Mtv call this one
  void add_mul_Mv_gen_sym_uptr(double vout[], const double vin[]); // M=M', then mul_Mtv call this one

  void mul_Mv_splitting(double vout[], const double vin[], int i_k);
  void mul_Mtv_splitting(double vout[], const double vin[], int i_k, int n_vout);
  void add_mul_Mv_splitting(double vout[], const double vin[], int i_k);
  void add_mul_Mtv_splitting(double vout[], const double vin[], int i_k, int n_vout);

  void mul_Mv_tridiag_shifted(double vout[], const double vin[]);
  void add_mul_Mv_tridiag_shifted(double vout[], const double vin[]);
  
  void mul_Mv_dense_diag_diag(double vout[], const double vin[]);
  void mul_Mtv_dense_diag_diag(double vout[], const double vin[]);
  void add_mul_Mv_dense_diag_diag(double vout[], const double vin[]);
  void add_mul_Mtv_dense_diag_diag(double vout[], const double vin[]);

  // Order a matrix packed format (both column-wise and row-wise)
  void order_packed_format(int begsize, int nz, int* &beg, int* &ind, double* &val);
  
 //! Auxiliary struct for sorting matrices in ija format.
  struct Order_ija {
    int *row; //!< Pointer to rows of matrix elements.
    int *col; //!< Pointer to columns of matrix elements.
    bool operator() (int i,int j) //! Comparison function
    {
      return ((row[i] < row[j]) || ((row[i] == row[j]) && (col[i] < col[j])));
    }
  };
  
  //! Auxiliary struct for sorting vectors.
  struct Order_vector {
    int *v; //!< Pointer to vector elements.
    bool operator() (int i,int j) //! Comparison function 
    {
      return (v[i] < v[j]);
    }
  };
};


inline void MatrixBlockIP::symbolic_fact_MMt(CHOL_SOLVER chslv, int *prov_pfa, int *prov_ipfa, NUMBERING prov_pfa_numbering)
//! Interface routine to SparseChol symbolic factorization
{
  chol.initialize(chslv,type,type_orientation);

  if (m <= 0) return; // for 0 rows matrices do nothing

  switch (type) {
  case GENERAL:
    if (!rowwise) native_to_general(false);
    chol.symbolic_fact_MMt(m, n, nz, icola, inirowa, a, num_blocks, prov_pfa, prov_ipfa, prov_pfa_numbering);
    break;
  case NETWORK:
    chol.symbolic_fact_MMt(num_nodes, num_arcs, src, dst, num_blocks, prov_pfa, prov_ipfa, prov_pfa_numbering);
    break;
  case IDENTITY:
  case IDTY_IDTY:
    chol.symbolic_fact_MMt(m, num_blocks);
    break;
  case DIAGONAL:
    chol.symbolic_fact_MMt(m,d1,num_blocks);
    break;
  case DIAG_DIAG:
    chol.symbolic_fact_MMt(m,d1,d2,num_blocks);
    break;
  case DENSE_DIAG_DIAG:
    chol.symbolic_fact_MMt(m,nd,num_blocks,ad,d1exist,d2exist,d1d,d2d,cd1,cd2);
    break;
  default:
    throw ExceptionBlockIP(NOT_FOR_THIS_MATRIX_TYPE, __FILE__, __LINE__);
  }
}

inline void MatrixBlockIP::numeric_fact_MMt(CHOL_SOLVER chslv, double *Theta, int i_k)
//! Interface routine to SparseChol numeric factorization
{
  if (m <= 0) return; // for 0 rows matrices do nothing

  // if first time then creation and symbolic factorization of M*Theta*M'
  if (!(chol.made_symbfct_MMt)) {
    symbolic_fact_MMt(chslv);
    chol.made_symbfct_MMt= true;
  }

  chol.numeric_fact_MMt(Theta, i_k);
}

inline void MatrixBlockIP::numeric_solve_MMt(double *rhs, int i_k, WHO_PERMUTES whoperm)
//! Interface routine to SparseChol numeric solve
{
  if (m <= 0) return; // for 0 rows matrices do nothing

  chol.numeric_solve_MMt(rhs, i_k, whoperm);
}

inline void MatrixBlockIP::symbolic_fact_M(CHOL_SOLVER chslv, int *prov_pfa, int *prov_ipfa, NUMBERING prov_pfa_numbering)
//! Interface routine to SparseChol symbolic factorization
{
  chol.initialize(chslv,type,type_orientation);

  if (m <= 0) return; // for 0 rows matrices do nothing

  // only implemented (and needed by now) for GEN_SYM_UPTR, DIAGONAL and TRIDIAG_SHIFTED matrices
  switch (type) {
  case DIAGONAL:
    chol.symbolic_fact_M(m);
    break;
  case IDENTITY:
    break;
  case GEN_SYM_UPTR:
    if (!rowwise) native_to_general(false);
    chol.symbolic_fact_M(m, nz, icola, inirowa, prov_pfa, prov_ipfa, prov_pfa_numbering);
    break;
  case TRIDIAG_SHIFTED:
    chol.symbolic_fact_M(m, t_shift);
    break;
  default:
    throw ExceptionBlockIP(NOT_FOR_THIS_MATRIX_TYPE, __FILE__, __LINE__);
  }
}

inline void MatrixBlockIP::numeric_fact_M(CHOL_SOLVER chslv)
//! Interface routine to SparseChol numeric factorization
{
  if (m <= 0) return; // for 0 rows matrices do nothing

  // if first time then symbolic factorization of M
  if (!(chol.made_symbfct_M)) {
    symbolic_fact_M(chslv);
    chol.made_symbfct_M= true;
  }

  switch (type) {
  case DIAGONAL:
    chol.numeric_fact_M(d1);
    break;
  case IDENTITY:
    // nothing to factorize
    break;
  case GEN_SYM_UPTR:
    chol.numeric_fact_M(a);
    break;
  case TRIDIAG_SHIFTED:
    chol.numeric_fact_M(tridiag_d,tridiag_s);
    break;
  default:
    throw ExceptionBlockIP(NOT_FOR_THIS_MATRIX_TYPE, __FILE__, __LINE__);
  }
}

inline void MatrixBlockIP::numeric_solve_M(double *rhs, WHO_PERMUTES whoperm)
//! Interface routine to SparseChol numeric solve
{
  if (m <= 0) return; // for 0 rows matrices do nothing

  chol.numeric_solve_M(rhs, whoperm);
}

inline int MatrixBlockIP::get_pfa(int i) 
//! Interface routine to SparseChol get_pfa(). 
/*! 
  @return pfa[i]
  @note The user must guarantee pfa previously computed in call to symbolic_fact().
 */
{
  return(chol.get_pfa(i));
}

inline int MatrixBlockIP::get_ipfa(int i)
//! Interface routine to SparseChol get_ipfa(). 
/*! 
  @return ipfa[i]
  @note The user must guarantee ipfa previously computed in call to symbolic_fact().
 */
{
  return(chol.get_ipfa(i));
}

inline int MatrixBlockIP::get_maxlnz()
//! Interface routine to SparseChol get_maxlnz(). 
/*! 
  @return maxlnz
  @note The user must guarantee maxlnz previously computed in call to symbolic_fact().
 */
{
  return (chol.get_maxlnz());
}

inline int MatrixBlockIP::get_maxfillin()
//! Interface routine to SparseChol get_maxfillin(). 
/*! 
  @return maxfillin
  @note The user must guarantee maxfillin previously computed in call to symbolic_fact().
 */
{
  return (chol.get_maxfillin());
}

inline int MatrixBlockIP::get_njka()
//! Interface routine to SparseChol get_njka(). 
/*! 
  @return njka
  @note The user must guarantee njka previously computed in call to symbolic_fact().
 */
{
  return (chol.get_njka());
}

inline int MatrixBlockIP::get_num_zero_pivots()
//! Interface routine to SparseChol get_num_zero_pivots().
/*! 
  @return num_zero_pivots
  @note It will 0 if no numeric factorization made.
 */
{
  return (chol.get_num_zero_pivots());
}

inline int MatrixBlockIP::get_num_semidef_matrix()
//! Interface routine to SparseChol get_semidef_matrix().
/*! 
  @return num_semidef_matrix
  @note It will 0 if no numeric factorization made.
 */
{
  return (chol.get_num_semidef_matrix());
}

inline void MatrixBlockIP::native_to_general(bool delete_native_format, int i_k, int n_k)
//! Calculates and stores the matrix in packed rowwise format
/*!
  @param delete_native_format If true only packed rowwise format will be stored
  @param i_k Only for case of SPLITTING matrix, index of block treated
  @param n_k Only for case of SPLITTING matrix, number of variables of the matrix (previously taken from N[i_k])
*/
{
  if (rowwise && type != SPLITTING) return; // already stored

  switch (type) {
  case GENERAL:
    if (columnwise)
      column_wise_to_row_wise_format();
    else if (ija)
      ija_to_rowwise();
    else {
      // this case should never happen; if it does, there is some internal error
      throw ExceptionBlockIP(NOIJA_GENERAL, __FILE__, __LINE__);
    }
    break;
  case NETWORK:
    network_to_general();
    break;
  case IDENTITY:
    identity_to_general();
    break;
  case DIAGONAL:
    diagonal_to_general();
    break;
  case IDTY_IDTY:
    idty_idty_to_general();
    break;
  case DIAG_DIAG:
    diag_diag_to_general();
    // Diego Juarez 05/15/2017. Added SPLITTING case
  case SPLITTING:
    splitting_to_general(i_k, n_k);
    // end modification Diego Juarez
    break;
  case DENSE_DIAG_DIAG:
    dense_diag_diag_to_general();
    break;
  default:
    throw ExceptionBlockIP(NOT_FOR_THIS_MATRIX_TYPE, __FILE__, __LINE__);
    break;
  }
  if (delete_native_format) {
    type= GENERAL;
    delete[] src;
    delete[] dst;
    delete[] d1;
    delete[] d2;
    src= NULL;
    dst= NULL;
    d1= NULL;
    d2= NULL;
  }
}

inline void MatrixBlockIP::identity_to_general()
//! Calculates and stores the I matrix in packed rowwise format
{
  restore();
  identity_to_ija_format();
  ija_to_rowwise();
}

inline void MatrixBlockIP::diagonal_to_general()
//! Calculates and stores the diagonal D matrix in packed rowwise format
{
  restore();
  diagonal_to_ija_format();
  ija_to_rowwise();
}

inline void MatrixBlockIP::idty_idty_to_general()
//! Calculates and stores the matrix [I I] in packed rowwise format
{
  restore();
  idty_idty_to_ija_format();
  ija_to_rowwise();
}

inline void MatrixBlockIP::diag_diag_to_general()
//! Calculates and stores the matrix [D1 D2] in packed rowwise format
{
  restore();
  diag_diag_to_ija_format();
  ija_to_rowwise();
}

inline void MatrixBlockIP::dense_diag_diag_to_general()
//! Calculates and stores the matrix [D D1 D2] in packed rowwise format
{
  restore();
  dense_diag_diag_to_ija_format();
  ija_to_rowwise();
}

//Diego Juarez 05/15/2017 - 05/16/2017. Add member function to produce a packed rowwise format from a splitting matrix
// Before this did not exists
inline void MatrixBlockIP::splitting_to_general(int i_k, int n_k)
//! Calculates and stores the i_k-th block of an splitting matrix in packed rowwise format
{
  rowwise = true; // Now is rowwise
  if (k_split == 1){ // One block means no splitting matrices
    delete[] inirowa; inirowa= NULL;
    inirowa= new int[m+1];
    icola = NULL;
    a = NULL;
    nz = 0;
    for (int i = 0; i<=m; ++i) inirowa[i]=0;
  }
  else{
    // Recompute the matrix just if previous computed block is different from i_k-th
    if(i_k_general_splitting != i_k){
      nz = (i_k == 0 || i_k==k_split-1) ? n_split : 2*n_split;
      // Allocate if there are no vectors yet 
      if(i_k_general_splitting == -1){
	inirowa = new int[m+1];
	icola = new int[nz];
	a = new double[nz];
      }
      // Or if we cannot use the same storage as the previous allocation
      else if (i_k != 0 && i_k != k_split -1 && (i_k_general_splitting == 0 || i_k_general_splitting == k_split-1) ||
	       (i_k == 0 || i_k == k_split-1) && i_k_general_splitting !=0 && i_k_general_splitting !=k_split-1){
	delete[] inirowa; inirowa= NULL;
	delete[] icola;   icola= NULL;
	delete[] a;       a= NULL;
	
	inirowa = new int[m+1];
	icola = new int[nz];
	a = new double[nz];
      }
      
      /*
	- If and only if ini_split <= column <= end_split, there will be a nonzero element:
        - If i_k == 0, then
	  - there will a +1 at position (column - ini_split, column).
	- If i_k == k_split - 1, then
	  - there will be a -1 at position (m - 1 - end_split + column, column).
	- Else,
	  - -1 at position [ (i_k-1)*n_split + column - ini_split, column];
	  - +1 at position [ i_k*n_split + column - ini_split, column].
    */
      
      inirowa[0]=0; // Always first row will be 0
      // i_k == 0
      if(i_k ==0){
	for(int i=1; i<=n_split; ++i) inirowa[i] = inirowa[i-1]+1;
	for(int i=n_split+1; i<=m; ++i) inirowa[i] = inirowa[i-1];
	
	for(int i=0; i<n_split; ++i) {
	  icola[i] = ini_split+i;
	  a[i] = +1;
	}
      }
      // i_k==k-1
      else if (i_k == k_split-1){
	for(int i=1; i<=(i_k-1)*n_split; ++i) inirowa[i]=inirowa[i-1];
	for(int i=(i_k-1)*n_split+1; i<=m; ++i) inirowa[i]=inirowa[i-1]+1;

	for(int i=0; i<n_split; ++i){
	  icola[i]=ini_split+i;
	  a[i] = -1;
	}	
      }
      else{
	for(int i=1; i<=(i_k-1)*n_split;++i) inirowa[i] = inirowa[i-1]; // First rows with zero
	for(int i=(i_k-1)*n_split+1;i<=(i_k+1)*n_split;++i) inirowa[i] = inirowa[i-1]+1; // Rows with non-zero elements
	for(int i=(i_k+1)*n_split+1;i<=m;++i) inirowa[i] = inirowa[i-1]; // Rest of rows with zero
	
	// Columns with -1
	for (int i=0; i<n_split; ++i){
	  icola[i]=ini_split+i;
	  a[i]= -1;
	}
	// Columns with +1
	for(int i=n_split; i<2*n_split; ++i){
	  icola[i]=i-n_split;
	  a[i]= +1;
	}
      }
      
      i_k_general_splitting = i_k; // Change last block computed
      n= n_k; // Set number of variables for this i_k-th submatrix of splitting matrix
    }
  }
}
// end modification Diego Juarez


inline void MatrixBlockIP::mul_Mv_row_wise(double vout[], const double vin[])
//! Matrix-vector product vout=M*vin (vout(m),vin(n)) (general rowwise matrix)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  int i,j;

  for(i=0;i<m;++i){
    vout[i]=0.0;
    for(j=inirowa[i];j<=inirowa[i+1]-1;++j) vout[i]+= a[j]*vin[icola[j]];
  }
}     

inline void MatrixBlockIP::mul_Mtv_row_wise(double vout[], const double vin[])
//! MatrixTranspose-vector product vout=M(t)*vin (vout(n),vin(m)  (general rowwise matrix)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{ 
  double r;
  int i,j;
  
  for(i=0;i<n;++i) vout[i]= 0.0;
  for(i=0;i<m;++i){
    r= vin[i];
    for(j=inirowa[i];j<=inirowa[i+1]-1;++j) vout[icola[j]]+= r*a[j];
  }
}   

inline void MatrixBlockIP::add_mul_Mv_row_wise(double vout[], const double vin[])
//! Add matrix-vector product vout += M*vin (vout(m),vin(n)) (general rowwise matrix)
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{
  int i,j;

  for(i=0;i<m;++i){
    for(j=inirowa[i];j<=inirowa[i+1]-1;++j) vout[i]+= a[j]*vin[icola[j]];
  }
}     

inline void MatrixBlockIP::add_mul_Mtv_row_wise(double vout[], const double vin[])
//! Add matrixTranspose-vector product vout += M(t)*vin (vout(n),vin(m)  (general rowwise matrix)
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{ 
  double r;
  int i,j;
  
  for(i=0;i<m;++i){
    r= vin[i];
    for(j=inirowa[i];j<=inirowa[i+1]-1;++j) vout[icola[j]]+= r*a[j];
  }
}

inline void MatrixBlockIP::mul_Mv_column_wise(double vout[], const double vin[])
//! Matrix-vector product vout=M*vin (vout(m),vin(n)) (general columnwise matrix)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  double r;
  int i,j;
  
  for(i=0;i<m;++i) vout[i]= 0.0;
  for(i=0;i<n;++i){
    r= vin[i];
    for(j=inicola[i];j<inicola[i+1];++j) vout[irowa[j]]+= r*a[j];
  }
}     

inline void MatrixBlockIP::mul_Mtv_column_wise(double vout[], const double vin[])
//! MatrixTranspose-vector product vout=M(t)*vin (vout(n),vin(m)  (general columnwise matrix)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{ 
  int i,j;

  for(i=0;i<n;++i){
    vout[i]=0.0;
    for(j=inicola[i];j<inicola[i+1];++j) vout[i]+= a[j]*vin[irowa[j]];
  }
}   

inline void MatrixBlockIP::add_mul_Mv_column_wise(double vout[], const double vin[])
//! Add matrix-vector product vout += M*vin (vout(m),vin(n)) (general columnwise matrix)
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{
  double r;
  int i,j;
  
  for(i=0;i<n;++i){
    r= vin[i];
    for(j=inicola[i];j<inicola[i+1];++j) vout[irowa[j]]+= r*a[j];
  }
}     

inline void MatrixBlockIP::add_mul_Mtv_column_wise(double vout[], const double vin[])
//! Add matrixTranspose-vector product vout += M(t)*vin (vout(n),vin(m)  (general columnwise matrix)
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{ 
  int i,j;

  for(i=0;i<n;++i){
    for(j=inicola[i];j<inicola[i+1];++j) vout[i]+= a[j]*vin[irowa[j]];
  }
}

inline void MatrixBlockIP::mul_Mv_network(double vout[], const double vin[])
//! Matrix-vector product vout=M*vin (vout(m),vin(n)) (network matrix, either oriented or nonoriented)
/*!
  @note It assumes vout has dimension m+1; last equation, related to last node, 
        is redundant and not used, but it is needed for an efficient computation.
  @note Element vout[m] is used (but backed-up at beginning and restored at ending)
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  double backvoutm= vout[m]; // fictitious position num_nodes backed-up

  if (type_orientation == ORIENTED) {
    // n= num_arcs, m= num_nodes-1
    int i;
    
    for(i=0;i<m+1;i++) vout[i]= 0; //this could be avoided if vout is already set to 0 by caller
    
    for(i=0;i<num_arcs;++i) {
      vout[src[i]]+= vin[i];
      vout[dst[i]]-= vin[i];
    }
  } 
  else { //NONORIENTED : matrix is [N -N]
    // n= 2*num_arcs, m= num_nodes-1
    // the following ordering of variables (arcs) is assumed:
    // first half of variables of vin[] associated to forward direction arcs;
    // second half of variables of vin[] associated to backward direction arcs
 
    int i;
    
    for(i=0;i<m+1;i++) vout[i]= 0; //this could be avoided if vout is already set to 0 by caller
    
    for(i=0;i<num_arcs;++i) {
      // forward arcs
      vout[src[i]]+= vin[i];
      vout[dst[i]]-= vin[i];
      // backward arcs
      vout[src[i]]-= vin[i+num_arcs];
      vout[dst[i]]+= vin[i+num_arcs];      
    }    
  }
  vout[m]= backvoutm; // fictitious position num_nodes restored
}

inline void MatrixBlockIP::mul_Mtv_network(double vout[], const double vin[])
//!  MatrixTranspose-vector product vout=M(t)*vin (vout(n),vin(m) (network matrix, either oriented or nonoriented)  
/*!
  @note It assumes vin has dimension m+1; last equation, related to last node, 
        is redundant and not used, but it is needed for an efficient computation.
  @note Element vin[m] is used (but backed-up at beginning and restored at ending)
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  double * vinback;
  vinback= const_cast<double *>(vin);
  double backvinm= vin[m]; // fictitious position num_nodes backed-up

  if (type_orientation == ORIENTED) {
    // n= num_arcs, m= num_nodes-1
    int i;
    
    vinback[m]= 0.0; // fictitious position num_nodes considered
    for(i=0;i<num_arcs;++i) {
      vout[i]= 0.0;
      vout[i]+= vin[src[i]];  // fictitious position num_nodes considered
      vout[i]-= vin[dst[i]];  // fictitious position num_nodes considered
    }
  } 
  else { //NONORIENTED : matrix is [N -N]
    // n= 2*num_arcs, m= num_nodes-1
    // the following ordering of variables (arcs) is assumed:
    // first half of vout[] associated to forward direction arcs;
    // second half of vout[] associated to backward direction arcs
    int i;
    
    vinback[m]= 0.0; // fictitious position num_nodes considered
    for(i=0;i<num_arcs;++i) {
      // this is for N*vin
      vout[i]= 0.0;
      vout[i]+= vin[src[i]];  // fictitious position num_nodes considered
      vout[i]-= vin[dst[i]];  // fictitious position num_nodes considered
      // this is for (-N)*vin: just neg vout[i] previously computed
      vout[i+num_arcs]= -vout[i]; 
    }
  }
  vinback[m]= backvinm; // fictitious position num_nodes restored
}

inline void MatrixBlockIP::add_mul_Mv_network(double vout[], const double vin[])
//! Add matrix-vector product vout += M*vin (vout(m),vin(n)) (network matrix, either oriented or nonoriented)
/*!
  @note It assumes vout has dimension m+1; last equation, related to last node, 
        is redundant and not used, but it is needed for an efficient computation.
  @note Element vout[m] is used (but backed-up at beginning and restored at ending)
  @param vout Result of the product
  @param vin Vector to product
*/
{
  double backvoutm= vout[m]; // fictitious position num_nodes backed-up

  if (type_orientation == ORIENTED) {
    // n= num_arcs, m= num_nodes-1
    int i;
    
   // caution: fictitious position num_nodes considered in vout
    for(i=0;i<num_arcs;++i) {
      vout[src[i]]+= vin[i];
      vout[dst[i]]-= vin[i];
    }
  } 
  else { //NONORIENTED: matrix is [N -N]
    // n= 2*num_arcs, m= num_nodes-1
    // the following ordering of variables (arcs) is assumed:
    // first half of variables of vin[] associated to forward direction arcs;
    // second half of variables of vin[] associated to backward direction arcs
 
    int i;
    
   // caution: fictitious position num_nodes considered in vout
    for(i=0;i<num_arcs;++i) {
      // forward arcs
      vout[src[i]]+= vin[i];
      vout[dst[i]]-= vin[i];
      // backward arcs
      vout[src[i]]-= vin[i+num_arcs];
      vout[dst[i]]+= vin[i+num_arcs];      
    }
  }
  vout[m]= backvoutm; // fictitious position num_nodes restored
}

inline void MatrixBlockIP::add_mul_Mtv_network(double vout[], const double vin[])
//! Add matrixTranspose-vector product vout += M(t)*vin (vout(n),vin(m) (network matrix, either oriented or nonoriented)
/*!
  @note It assumes vin has dimension m+1; last equation, related to last node, 
        is redundant and not used, but it is needed for an efficient computation.
  @note Element vin[m] is used (but backed-up at beginning and restored at ending)
  @param vout Result of the product
  @param vin Vector to product
*/
{
  double * vinback;
  vinback= const_cast<double *>(vin);
  double backvinm= vin[m]; // fictitious position num_nodes backed-up

  if (type_orientation == ORIENTED) {
    // n= num_arcs, m= num_nodes-1
    int i;
    
    vinback[m]= 0.0; // fictitious position num_nodes considered
    for(i=0;i<num_arcs;++i) {
      vout[i]+= vin[src[i]];  // fictitious position num_nodes considered
      vout[i]-= vin[dst[i]];  // fictitious position num_nodes considered
    }
  } 
  else { //NONORIENTED: matrix is [N -N]
    // n= 2*num_arcs, m= num_nodes-1
    // the following ordering of variables (arcs) is assumed:
    // first half of vout[] associated to forward direction arcs;
    // second half of vout[] associated to backward direction arcs
    int i;
    
    vinback[m]= 0.0; // fictitious position num_nodes considered
    for(i=0;i<num_arcs;++i) {
      double raux= vin[src[i]]-vin[dst[i]]; // fictitious position num_nodes considered
      // this is for N*vin
      vout[i]+= raux;
      // this is for (-N)*vin
      vout[i+num_arcs]= -raux; 
    }
  }
  vinback[m]= backvinm; // fictitious position num_nodes restored
}

inline void MatrixBlockIP::mul_Mv_identity(double vout[], const double vin[])
//! Identity-vector product vout=M*vin (vout(m),vin(n))
/*!
  Identity matrix, m=n, just copy input to output vector
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  for (int i= 0; i < n; ++i) vout[i]= vin[i];
}

inline void MatrixBlockIP::add_mul_Mv_identity(double vout[], const double vin[])
//! Add Identity-vector product vout += M*vin (vout(m),vin(n))
/*!
  Identity matrix, m=n, just add input to output vector
  @param vout Result of the product
  @param vin Vector to product
*/
{
  for (int i= 0; i < n; ++i) vout[i]+= vin[i];
}

inline void MatrixBlockIP::mul_Mv_diagonal(double vout[], const double vin[])
//! Diagonal-vector product vout=D*vin (vout(m),vin(n)) D is diagonal, m=n
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  for (int i= 0; i < n; ++i) vout[i]= d1[i]*vin[i];
}


inline void MatrixBlockIP::add_mul_Mv_diagonal(double vout[], const double vin[])
//! Add Diagonal-vector product vout += D*vin (vout(m),vin(n)) D is diagonal, m=n
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{
  for (int i= 0; i < n; ++i) vout[i]+= d1[i]*vin[i];
}


inline void MatrixBlockIP::mul_Mv_idty_idty(double vout[], const double vin[])
//! Matrix-vector product vout= [I I]*vin (vout(m),vin(n)) (M= [I I] matrix, n= 2*m)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  for (int i= 0, j=m; i<m; i++, j++) vout[i]= vin[i]+vin[j];
}     

inline void MatrixBlockIP::mul_Mtv_idty_idty(double vout[], const double vin[])
//! MatrixTranspose-vector product vout= [I I]'*vin (vout(n),vin(m)   (M= [I I] matrix, n= 2*m)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{ 
  for (int i= 0, j=m; i<m; i++, j++) {
    vout[i]= vin[i];
    vout[j]= vin[i];
  }
}   

inline void MatrixBlockIP::add_mul_Mv_idty_idty(double vout[], const double vin[])
//! Add matrix-vector product vout += [I I]*vin (vout(m),vin(n))  (M= [I I] matrix, n= 2*m)
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{
  for (int i= 0, j=m; i<m; i++, j++) vout[i]+= vin[i]+vin[j];
}     

inline void MatrixBlockIP::add_mul_Mtv_idty_idty(double vout[], const double vin[])
//! Add matrixTranspose-vector product vout += [I I]'*vin (vout(n),vin(m)  (M= [I I] matrix, n= 2*m)
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{ 
  for (int i= 0, j=m; i<m; i++, j++) {
    vout[i]+= vin[i];
    vout[j]+= vin[i];
  }
}   

inline void MatrixBlockIP::mul_Mv_diag_diag(double vout[], const double vin[])
//! Matrix-vector product vout= [D1 D2]*vin (vout(m),vin(n)) (M= [D1 D2] matrix, n= 2*m)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  for (int i= 0, j=m; i<m; i++, j++) vout[i]= d1[i]*vin[i]+d2[i]*vin[j];
}     

inline void MatrixBlockIP::mul_Mtv_diag_diag(double vout[], const double vin[])
//! MatrixTranspose-vector product vout= [D1 D2]'*vin (vout(n),vin(m)   (M= [D1 D2] matrix, n= 2*m)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{ 
  for (int i= 0, j=m; i<m; i++, j++) {
    vout[i]= d1[i]*vin[i];
    vout[j]= d2[i]*vin[i];
  }
}   

inline void MatrixBlockIP::add_mul_Mv_diag_diag(double vout[], const double vin[])
//! Add matrix-vector product vout += [D1 D2]*vin (vout(m),vin(n))  (M= [D1 D2] matrix, n= 2*m)
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{
  for (int i= 0, j=m; i<m; i++, j++) vout[i]+= d1[i]*vin[i]+d2[i]*vin[j];
}     

inline void MatrixBlockIP::add_mul_Mtv_diag_diag(double vout[], const double vin[])
//! Add matrixTranspose-vector product vout += [D1 D2]'*vin (vout(n),vin(m)  (M= [D1 D2] matrix, n= 2*m)
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{ 
  for (int i= 0, j=m; i<m; i++, j++) {
    vout[i]+= d1[i]*vin[i];
    vout[j]+= d2[i]*vin[i];
  }
}

inline void MatrixBlockIP::mul_Mv_gen_sym_uptr(double vout[], const double vin[])
//! Matrix-vector product vout=M*vin (vout(m),vin(n)) (general symmetric upper triangular rowwise matrix)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  for(int i=0;i<m;i++) vout[i]= 0.0; 
  for(int i=0;i<m;i++){
    for(int l=inirowa[i];l<=inirowa[i+1]-1;l++) {
      int j= icola[l];
      vout[i]+= a[l]*vin[j]; // vout[i] += a[i,j]*vin[j]
      if (i != j) vout[j]+= a[l]*vin[i]; // if upper-(non-)diagonal element then vout[j] += a[j,i]*vin[i]
    }
  }
}     

inline void MatrixBlockIP::add_mul_Mv_gen_sym_uptr(double vout[], const double vin[])
//! Add matrix-vector product vout += M*vin (vout(m),vin(n)) (general symmetric upper triangular rowwise matrix)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  for(int i=0;i<m;i++){
    for(int l=inirowa[i];l<=inirowa[i+1]-1;l++) {
      int j= icola[l];
      vout[i]+= a[l]*vin[j]; // vout[i] += a[i,j]*vin[j]
      if (i != j) vout[j]+= a[l]*vin[i]; // if upper-(non-)diagonal element then vout[j] += a[j,i]*vin[i]
    }
  }
}     

inline void MatrixBlockIP::mul_Mv_splitting(double vout[], const double vin[], int i_k)
//! Matrix-vector product vout= SPL*vin (vout(m),vin(n)) (SPL= splitting matrix)
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
  @param i_k Block to be used 
*/
{
  /*
    for example, for a k=4 blocks splitting matrix we have
          |0 I 0   0-I 0  0 0 0  0 0 0 |
     L =  |0 0 0   0 I 0  0-I 0  0 0 0 | =  [L_1 L_2 L_3 L_4]
          |0 0 0   0 0 0  0 I 0  0-I 0 |
   */

  if (m==0) return;

  if (i_k==0) { //first block
    for(int j=0, i=ini_split ; i<=end_split ; i++,j++) vout[j]= vin[i];
    for(int j=n_split ; j<m ; j++) vout[j]= 0.0;
  }
  else if (i_k==k_split-1) { //last block
    for(int j=0 ; j<m-n_split ; j++) vout[j]= 0.0;
    for(int j= m-n_split, i=ini_split ; i<=end_split ; i++,j++) vout[j]= -vin[i];
  }
  else{ //mid blocks
    // with minor changes this code would include previous cases, 
    // however we keep them separated for clarity
    for(int j=0 ; j<(i_k-1)*n_split ; j++) vout[j]= 0.0;
    for(int j=(i_k-1)*n_split, i=ini_split ; i<=end_split ; i++,j++) vout[j]= -vin[i];
    for(int j=i_k*n_split, i=ini_split ; i<=end_split ; i++,j++) vout[j]= vin[i];
    for(int j=(i_k+1)*n_split ; j<m ; j++) vout[j]= 0.0;
  }
}     

inline void MatrixBlockIP::mul_Mtv_splitting(double vout[], const double vin[], int i_k, int n_vout)
//! MatrixTranspose-vector product vout= SPL'*vin (vout(n),vin(m)   (SPL= splitting matrix))
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
  @param i_k Block to be used 
  @param n_vout Size of vout (it is unknown for splitting matrices)
*/
{ 
  /*
    for example, for a k=4 blocks splitting matrix we have

          | 0  0  0|
          | I  0  0|
	  | 0  0  0|
          | 	   |
          | 0  0  0|
	  |-I  I  0|   | L1' |
	  | 0  0  0|   | L2' |
    L'=	  |	   | = | L3' |
          | 0  0  0|   | L4' |
	  | 0- I  I|   
	  | 0  0  0|
	  |	   |
          | 0  0  0|
	  | 0  0 -I|
	  | 0  0  0|
   */
  
  if (m==0) return;
  
  if (i_k==0) { //first block
    for(int j=0 ; j<ini_split ; j++) vout[j]= 0.0;
    for(int j=ini_split,i=0 ; j<end_split+1 ; j++, i++) vout[j]= vin[i];
    for(int j=end_split+1 ; j<n_vout; j++) vout[j]= 0.0;
  }
  else if (i_k==k_split-1) { //last block
    for(int j=0 ; j<ini_split ; j++) vout[j]= 0.0;
    for(int j=ini_split,i=m-n_split ; j<end_split+1 ; j++, i++) vout[j]= -vin[i];
    for(int j=end_split+1 ; j<n_vout; j++) vout[j]= 0.0;
  }
  else{ //mid blocks
    // with minor changes this code would include previous cases, 
    // however we keep them separated for clarity
    for(int j=0 ; j<ini_split ; j++) vout[j]= 0.0;
    for(int j=ini_split,i=(i_k-1)*n_split,k=i_k*n_split ; j<end_split+1 ; j++, i++, k++) vout[j]= -vin[i]+vin[k];
    for(int j=end_split+1 ; j<n_vout; j++) vout[j]= 0.0;
  }
}       

inline void MatrixBlockIP::add_mul_Mv_splitting(double vout[], const double vin[], int i_k)
//! Add matrix-vector product vout += SPL*vin (vout(m),vin(n))  (SPL= splitting matrix)
/*!
  @param vout Result of the product
  @param vin Vector to product
  @param i_k Block to be used 
*/
{
  /*
    for example, for a k=4 blocks splitting matrix we have
          |0 I 0   0-I 0  0 0 0  0 0 0 |
     L =  |0 0 0   0 I 0  0-I 0  0 0 0 | =  [L_1 L_2 L_3 L_4]
          |0 0 0   0 0 0  0 I 0  0-I 0 |
   */
 
  if (m==0) return;

  if (i_k==0) { //first block
    for(int j=0, i=ini_split ; i<=end_split ; i++,j++) vout[j] += vin[i];
  }
  else if (i_k==k_split-1) { //last block
    for(int j= m-n_split, i=ini_split ; i<=end_split ; i++,j++) vout[j] += -vin[i];
  }
  else{ //mid blocks
    // with minor changes this code would include previous cases, 
    // however we keep them separated for clarity
    for(int j=(i_k-1)*n_split, i=ini_split ; i<=end_split ; i++,j++) vout[j] += -vin[i];
    for(int j=i_k*n_split, i=ini_split ; i<=end_split ; i++,j++) vout[j] += vin[i];
  }
}     

inline void MatrixBlockIP::add_mul_Mtv_splitting(double vout[], const double vin[], int i_k, int n_vout)
//! Add matrixTranspose-vector product vout += SPL'*vin (vout(n),vin(m)  (SPL= splitting matrix)
/*!
  @param vout Result of the product
  @param vin Vector to product
  @param i_k Block to be used 
  @param n_vout  Size of vout (it is unknown for splitting matrices)
*/
{ 
  /*
    for example, for a k=4 blocks splitting matrix we have

          | 0  0  0|
          | I  0  0|
	  | 0  0  0|
          | 	   |
          | 0  0  0|
	  |-I  I  0|   | L1' |
	  | 0  0  0|   | L2' |
    L'=	  |	   | = | L3' |
          | 0  0  0|   | L4' |
	  | 0- I  I|   
	  | 0  0  0|
	  |	   |
          | 0  0  0|
	  | 0  0 -I|
	  | 0  0  0|
   */

  if (m==0) return;

  if (i_k==0) { //first block
    for(int j=ini_split,i=0 ; j<end_split+1 ; j++, i++) vout[j] += vin[i];
  }
  else if (i_k==k_split-1) { //last block
    for(int j=ini_split,i=m-n_split ; j<end_split+1 ; j++, i++) vout[j] += -vin[i];
  }
  else{ //mid blocks
    // with minor changes this code would include previous cases, 
    // however we keep them separated for clarity
    for(int j=ini_split,i=(i_k-1)*n_split,k=i_k*n_split ; j<end_split+1 ; j++, i++, k++) vout[j] += (-vin[i]+vin[k]);
  }

}   

inline void MatrixBlockIP::mul_Mv_tridiag_shifted(double vout[], const double vin[])
//! Diagonal-vector product vout=T*vin (vout(m),vin(n)) T is tridiagonal shifted matrix, m=n
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{
  /*
    Matrix is symmetric with main diagonal tridiag_d and lower/upper-diagonal tridiag_s; 
    both main and lower/upper diagonal separated t_shift positions. For example, for m=n=9 and
    t_shift= 3

     1   2   3   4   5    6   7  8  9
                                                    (a)    (b)    (c)
 1  |d1          s1                   |  |v1|    |        d1v1 + s1v4 |
 2  |    d2          s2               |  |v2|    |        d2v2 + s2v5 |
 3  |        d3          s3           |  |v3|    |        d3v3 + s3v6 |
 4  |s1          d4          s4       |  |v4| =  | s1v1 + d4v4 + s4v7 |
 5  |    s2          d5          s5   |  |v5|    | s2v2 + d5v5 + s5v8 |
 6  |        s3          d6         s6|  |v6|    | s3v3 + d6v6 + s6v9 |
 7  |            s4          d7       |  |v7|    | s4v4 + d7v7        |
 8  |                s5          d8   |  |v8|    | s5v5 + d8v8        |
 9  |                    s6         d9|  |v9|    | s6v6 + d9v9        |

  */
  
  for (int i=0;i<m;i++) vout[i]= tridiag_d[i]*vin[i]; // part (b)
  for (int i=0, j=t_shift;i<m-t_shift;i++, j++) vout[i] += tridiag_s[i]*vin[j]; // part (c)
  for (int i=0, j=t_shift;i<m-t_shift;i++, j++) vout[j] += tridiag_s[i]*vin[i]; // part (a)
}


inline void MatrixBlockIP::add_mul_Mv_tridiag_shifted(double vout[], const double vin[])
//! Add Diagonal-vector product vout += T*vin (vout(m),vin(n)) T is tridiagonal shifted matrix, m=n
/*!
  @param vout Result of the product
  @param vin Vector to product
*/
{
  /*
    See explanation of previous routine mul_Mv_tridiag_shifted()
  */
  
  for (int i=0;i<m;i++) vout[i]+= tridiag_d[i]*vin[i]; // part (b)
  for (int i=0, j=t_shift;i<m-t_shift;i++, j++) vout[i] += tridiag_s[i]*vin[j]; // part (c)
  for (int i=0, j=t_shift;i<m-t_shift;i++, j++) vout[j] += tridiag_s[i]*vin[i]; // part (a)
}

inline void MatrixBlockIP::mul_Mv_dense_diag_diag(double vout[], const double vin[])
//! Matrix-vector product vout= [D D1 D2]*vin (vout(m),vin(n))
/*!
  @param vout Result of the product. Previously allocated
  @param vin Vector to product
*/
{

  // Dense part
  for(int i=0;i<m;i++) vout[i]= 0.0;
  for(int j=0;j<nd;j++) {
    double raux= vin[j];
    double *colAj= &(ad[j*m]);
    for(int i=0;i<m;i++) vout[i] += raux*colAj[i];
  }

  // D1 and D2 parts 
  if (d1exist) {
    const double *vin_d1= &(vin[nd]);
    if (cd1==0.0)
      for(int i=0;i<m;i++) vout[i] += d1d[i]*vin_d1[i];
    else
      for(int i=0;i<m;i++) vout[i] += cd1*vin_d1[i];
    if (d2exist) {
      const double *vin_d2= &(vin[nd+m]);
      if (cd2==0.0)
	for(int i=0;i<m;i++) vout[i] += d2d[i]*vin_d2[i];
      else
	for(int i=0;i<m;i++) vout[i] += cd2*vin_d2[i];
    }
  }
}     

inline void MatrixBlockIP::mul_Mtv_dense_diag_diag(double vout[], const double vin[])
//! MatrixTranspose-vector product vout= [D D1 D2]'*vin (vout(n),vin(m)   
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{ 
  // Dense part vout(1:nd)=D'*vin
  for(int j=0;j<nd;j++) {
    double raux= 0.0;
    double *colAj= &(ad[j*m]);
    for(int i=0;i<m;i++) raux +=  vin[i]*colAj[i];
    vout[j]= raux;
  }

  // D1  and D2
  if (d1exist) {
    double *vout_d1= &(vout[nd]);
    if (cd1==0.0)
      for (int i= 0; i<m; i++) vout_d1[i]= d1d[i]*vin[i];
    else 
      for (int i= 0; i<m; i++) vout_d1[i]= cd1*vin[i];
    if (d2exist) {
      double *vout_d2= &(vout[nd+m]);
      if (cd2==0.0)
	for (int i= 0; i<m; i++) vout_d2[i]= d2d[i]*vin[i];
      else 
	for (int i= 0; i<m; i++) vout_d2[i]= cd2*vin[i];
    }
  }
}   

inline void MatrixBlockIP::add_mul_Mv_dense_diag_diag(double vout[], const double vin[])
//! Add matrix-vector product vout += [D D1 D2]*vin (vout(m),vin(n))
/*!
  @param vout Result of the product. Previously allocated
  @param vin Vector to product
*/
{

  // Dense part
  for(int j=0;j<nd;j++) {
    double raux= vin[j];
    double *colAj= &(ad[j*m]);
    for(int i=0;i<m;i++) vout[i] += raux*colAj[i];
  }

  // D1 and D2 parts 
  if (d1exist) {
    const double *vin_d1= &(vin[nd]);
    if (cd1==0.0)
      for(int i=0;i<m;i++) vout[i] += d1d[i]*vin_d1[i];
    else
      for(int i=0;i<m;i++) vout[i] += cd1*vin_d1[i];
    if (d2exist) {
      const double *vin_d2= &(vin[nd+m]);
      if (cd2==0.0)
	for(int i=0;i<m;i++) vout[i] += d2d[i]*vin_d2[i];
      else
	for(int i=0;i<m;i++) vout[i] += cd2*vin_d2[i];
    }
  }
}     

inline void MatrixBlockIP::add_mul_Mtv_dense_diag_diag(double vout[], const double vin[])
//! Add matrixTranspose-vector product vout += [D D1 D2]'*vin (vout(n),vin(m)   
/*!
  @param vout Result of the product. Must be allocated
  @param vin Vector to product
*/
{ 
  // Dense part vout(1:nd)=D'*vin
  for(int j=0;j<nd;j++) {
    double raux= 0.0;
    double *colAj= &(ad[j*m]);
    for(int i=0;i<m;i++) raux +=  vin[i]*colAj[i];
    vout[j] += raux;
  }

  // D1  and D2
  if (d1exist) {
    double *vout_d1= &(vout[nd]);
    if (cd1==0.0)
      for (int i= 0; i<m; i++) vout_d1[i] += d1d[i]*vin[i];
    else 
      for (int i= 0; i<m; i++) vout_d1[i] += cd1*vin[i];
    if (d2exist) {
      double *vout_d2= &(vout[nd+m]);
      if (cd2==0.0)
	for (int i= 0; i<m; i++) vout_d2[i] += d2d[i]*vin[i];
      else 
	for (int i= 0; i<m; i++) vout_d2[i] += cd2*vin[i];
    }
  }
}   

inline void MatrixBlockIP::compute_D(int numBlocks, bool sameL, MatrixBlockIP L[], bool isActive[], int iniTheta[], double Theta[])
//! Compute D = Theta_(numBlocks+1) + sum{i..numBlocks} L_i*Theta_i*L_i'
/*!
  @param numBlocks Number of diagonal blocks
  @param sameL Define if the same matrix is used for each L block
  @param isActive Says what linking constrains are active
  @param iniTheta Indice to the first element of the Theta matrix of each block
  @param Theta numBlocks diagonal matrices with dimension n_i x n_i
*/
{
  if (!made_analyze_D) {
    analyze_D(numBlocks,sameL,L);
    made_analyze_D= true;
  }

  if (type==DIAGONAL) 
    compute_D_diagonal(numBlocks,sameL,L,isActive,iniTheta,Theta);
  else if (type==GEN_SYM_UPTR) 
    compute_D_gen_sym_uptr(numBlocks,sameL,L,isActive,iniTheta,Theta);
  else if (type==TRIDIAG_SHIFTED)
    compute_D_tridiag_shifted(numBlocks,sameL,L,isActive,iniTheta,Theta);
}

inline void MatrixBlockIP::analyze_D(int numBlocks, bool sameL, MatrixBlockIP L[])
//! Create internal structure arrays to compute D = Theta_(numBlocks+1) + sum{i..numBlocks} L_i*Theta_i*L_i'
//! It decides whether D is DIAGONAL or GEN_SYM_UPTR
/*!
  @param numBlocks Number of diagonal blocks
  @param sameL Define if the same matrix is used for each L block
  @param L Linking constraints blocks
*/
{
  reset();
  m= L[0].m; // All L_i have the same number of rows
  n= m;
  
  int sizeL= sameL ? 1 : numBlocks;
  bool is_diag= true;
  for(int i=0 ; i<sizeL && is_diag ;i++) {
    if (!(L[i].type == IDENTITY || L[i].type== IDTY_IDTY || 
	  L[i].type== DIAGONAL  || L[i].type== DIAG_DIAG )) is_diag= false;
  }
  if (is_diag) {
    // D is diagonal positive semidefinite matrix
    type= DIAGONAL;
    analyze_D_diagonal(numBlocks, sameL, L);
  }
  else if (L[0].type==SPLITTING) {
    // if L is SPLITTING then L*L' is TRIDIAGONAL_SHIFTED matrix
    type= TRIDIAG_SHIFTED;
    analyze_D_tridiag_shifted(numBlocks, sameL, L);
  }
  else {
    // D is symmetric positive semidefinite matrix 
    type= GEN_SYM_UPTR;
    rowwise= true;
    analyze_D_gen_sym_uptr(numBlocks, sameL, L);
  }
}


#endif //MATRIXBLOCKIP_H
