#ifndef SPCHOL_H
#define SPCHOL_H

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include "enums.h"
#include "ExceptionBlockIP.h"

using namespace std;

//! Class for sparse Cholesky factorizations
class SparseChol {
    
  // data representation

 public:
   /*! boolean to check whether symbolic factorization already made */
  bool made_symbfct_MMt; // for symbolic factorization of M*Theta*M'
  bool made_symbfct_M;   // for symbolic factorization of M
  
 private:

  //! Auxiliary struct for sorting network structure.
  typedef struct { 
    int src; //!< Source node of arc.
    int dst; //!< Destination node of arc
    int arc; //!< Arc.
  } SRC_DST_ARC;

  CHOL_SOLVER chol_solver; //!< Cholesky solver to be used.
  TYPE_MATRIX type_matrix; //!< Type of matrix (general, network, identity, etc),.
  TYPE_ORIENTATION type_orientation; //!< Type of arc orientation for network matrices (by default, oriented).

  //! Threshold pivot value for Cholesky factorizations of "simple" (diag, idty...) matrices (no pivot less than or equal to MIN_PIVOT allowed) 
  static constexpr double MIN_PIVOT= 0.0; 
  //! New value for pivots below MIN_PIVOT in Cholesky factorizations of "simple" matrices (diag, idty...)
  static constexpr double POSITIVE_PIVOT= 1.0E+128; 
  
  int m; //!< Dimension of system A*A'/A : m x m matrix.
  int njka; //!< Nonzeros in diagonal and subdiagonal of A*A'/A
  int maxfillin; //!< Nonzeros in factorization of A*A'/A due to fill-in.
  int maxlnz; //!< Total nonzeros in factorization of A*A'/A (= maxfillin+njka).
  int k; //!< Number of matrices A*A' with same topology to be factorized, see below explanation of lnz() and pnlz() (k=1 for A)
  
  int num_zero_pivots; //!< Number of zero pivots found during factorizations.
  int num_semidef_matrix; //!< Number of semidefinite matrices found during factorizations.

  
  // for diagonal matrices
  double *d1; //!< Diagonal of DIAG or first diagonal of DIAG_DIAG [D1 D2] matrices.
  double *d2; //!< Second diagonal of DIAG_DIAG [D1 D2] matrices.

  // for tridiag_shifted matrices
  int t_shift; //!< distance between diagonal and non-diagonal vectors 

  // for dense_diag_diag matrices [D D1 D2]
  int nd; //!< columns of dense matrix
  bool d1exist, d2exist; //!< whether therar are D1 and D2 matrices
  double cd1, cd2; //!< constant value if D1 and D2 are constant
  double *ad;  //!< pointer to dense part D, copied from MatrixBlockIP class, not to be deleted by SparsChol
  double *d1d1, *d2d2; //!< diagonal matrices D1*D1 and D2*D2
  

  /** @name  
      Vectors related to symmetric row-column permutation:
  */
  //@{
  /** Vector pfa of direct row permutation of matrix A*A', and its inverse (ipfa) computed 
      by Cholesky solver.
      
      pfa[i]= k : original row k is now (after permutation) in position i.\n
      ipfa[k]= i : i is current row (after permutation) of original row k.\n
      Both pfa and ipfa needed  because permutation may be nonsymmetric.\n
      pfa_numbering for either C-0 or Fortran-1 numbering style.\n
      permrhs(): auxiliary array, stores the permutation of the rhs when solving the system.
  */
  int *pfa; //!< pfa[i]= k : original row k is now (after permutation) in position i
  int *ipfa; //!< ipfa[k]= i : i is current row (after permutation) of original row k.
  NUMBERING pfa_numbering; //!< Numbering style of pfa (used in C and Fortran code).
  double *permrhs; //!< Auxliary array to store the permutation of the rhs of the system to be solved.
  //@}

  /** @name  
      When chol_solver is sprsblkllt:
  */

  /** @name  
      When chol_solver is sprsblkllt and A is 'GENERAL' matrix :
   */
  //@{
  /** When A is 'GENERAL' matrix: \n
    ia,ja,ka,nla,la and va are internal indices for efficiently computing A*Theta*A' and
    symbolic factorization using the procedure
    described in  "Monma, C. L. and A.J. Morton. 1987. Computational experience 
    with a dual affine variant of Karmarkar's method for linear programming.
    Operations Research Letters V.6 n.6".\n
    
    ia(m+1): pointer to first in ja(.)\n
    ja(njka): pair of rows (ia,ja) of A that generate arc in graph of A*A' \n
    ka(njka+1): pointer to first in la(.) and va(.) \n
    nla: size of la() and va() \n
    la(nla): columns that intervene in arc (i,j) in graph of A*A' \n
    va(nla): premultiplied values associated to column of la(.) \n
  */

  int *ia; //!< ia(m+1): pointer to first in ja().
  int *ja; //!< ja(njka): pair of rows (ia,ja) of A that generate arc in graph of A*A' .
  int *ka; //!< ka(njka+1): pointer to first in la() and va().
  int *la; //!< la(nla): columns that intervene in arc (i,j) in graph of A*A'.
  int nla; //!< Size of la() and va().
  double *va; //!< va(nla): premultiplied values associated to column of la().
  //@}

  /** @name  
      When chol_solver is sprsblkllt and A is 'NETWORK' matrix:
  */
  //@{
  /** 
      When A is 'NETWORK' matrix: \n
      nnu, nar, *inik, *dst2, *arck_l, *inil, *src2, *arcl_k are
      internal indices and variables for efficiently computing A*Theta*A'
      and symbolic factorization.       
  */
  int nnu; //!< Number of nodes.
  int nar; //!< Number of arcs.
  int *inik; //!< inik(nnu+1): pointers to first arc sorted by source (compressed form).
  int *dst2; //!< dst2(nar): destination nodes of arcs in inik.
  int *arck_l; //!< arck_l(nar): arcs of inik, dst2.
  int *inil; //!< inil(nnu+1): pointers to first arc sorted by destination (compressed form).
  int *src2; //!< src2(nar): source nodes of arcs in inik
  int *arcl_k; //!< arcl_k(nar): arcs of inil, src2.
  //@}
 
 
  /** @name  
      Variables needed by sprsblkllt:
  */
  //@{
  /*!
    Variables needed by sprsblkllt:
    
    xadj(m+1): indices to adjacency matrix ajcncy().\n
    ajcncy(2*(njka-m)): 2*arcs in graph of A*A'/A without diagonal entries (i,i) of A*A'/A.\n
    colnct(m): elements per column in factorization.\n
    snode(m): supernode of each column.\n
    xsuper(m+1): pointer to first column of supernode.\n
    xlindx(maxsuper): indices to lindx (compressed form).\n
    lindx(maxsub): compressed form of sub and diagonal of factorization.\n
    split(m): splitting of supernodes for exploiting cache memory.\n
    xlnz(m+1): indices to first column/row in lnz(.).\n
    lnz(k*maxlnz): sub(upper) and  diagonal elements of  ATheta(k)A' (and its Cholesky factorization, once performed).\n
    plnz(k): pointer to lnz(i); this is used when k matrices with same structure must be factorized; their share symbolic factorization, thes only difference is in lnz(); we need a different lnz() for each of them; i=0..k-1, so first matrix is matrix 0, last matrix is k-1.\n
  */
  int *xadj; //!< xadj(m+1): indices to adjacency matrix ajcncy().
  int *ajcncy; //!< ajcncy(2*(njka-m)): 2*arcs in graph of A*A'/A without diagonal entries (i,i) of A*A'/A.
  int *workspak; //!< 
  int *xlnz; //!< xlnz(m+1): indices to first column/row in lnz(.).
  int *xlindx; //!< xlindx(maxsuper): indices to lindx (compressed form).
  int *lindx; //!< lindx(maxsub): compressed form of sub and diagonal of factorization.
  int maxsub; //!< Maximum size of some vectors, computed by sprsblkllt.
  int *colcnt; //!< colnct(m): elements per column in factorization.
  int *snode; //!< snode(m): supernode of each column.
  int maxsuper; //!< Maximum size of some vectors, computed by sprsblkllt.
  int *xsuper; //!< xsuper(m+1): pointer to first column of supernode.
  int iwsiz; //!< Size of integer working space, required by sprsblkllt.
  int *split;//!< split(m): splitting of supernodes for exploiting cache memory.
  int tmpsiz; //!< Size of temporary working space, required by sprsblkllt.
  double *lnz; //!< lnz(k*maxlnz): sub(upper) and  diagonal elements of  ATheta(k)A' (and its Cholesky factorization, once performed).
  //
  //! plnz(k): pointers to lnz(i).
  /*! This is used when k matrices with same structure must be factorized; 
    their share symbolic factorization, the only difference is in lnz(); we need a different lnz() for 
    each of them; i=0..k-1, so first matrix is matrix 0, last matrix is k-1. 
  */
  double **plnz;
  //@}

  /** @name  
      Variables with indices for fast filling of matrix to factorize:
  */
  //@{
  /*! Variables with indices for fast filling of lnz().\n 
    If A is 'GENERAL' matrix:\n
    ilnz(njka): indices to lnz(.) of each nonzero element of A*Theta*A'/A.\n
    ifillin(maxfillin): fill-in positions in lnz(.) to be cleaned before filling lnz() with A*Theta*A'.\n
    \n
    If A is 'NETWORK' matrix:\n
    iarc(maxiarc): indices to arcs intervening in each element nonzero of lnz().\n
    maxiarc: size of iarc.\n
    ini_arc(njka+1): begin of arcs in iarc for each element of lnz().\n
  */
  int *ilnz; //!< ilnz(njka): indices to lnz(.) of each nonzero element of A*Theta*A'.
  int *ifillin; //!< ifillin(maxfillin): fill-in positions in lnz(.) to be cleaned before filling lnz() with A*Theta*A'.
  int maxiarc; //!< Size of iarc.
  int *iarc; //!< iarc(maxiarc): indices to arcs intervening in each element nonzero of lnz().
  int *ini_arc; //!< ini_arc(njka+1): begin of arcs in iarc for each element of lnz().
  //@}
  
  // ********** if chol_solver is cholmod:
  // not yet implemented


  // member functions
    
public: //interface routines

    SparseChol(); // Constructor
    ~SparseChol(); // Destructor
    void reset(CHOL_SOLVER chol_solver= (CHOL_SOLVER)NULL, TYPE_MATRIX type_matrix= (TYPE_MATRIX)NULL, 
	       TYPE_ORIENTATION type_orientation= ORIENTED); // frees memory and set to 0 (like calling destructor+constructor)
    void initialize(CHOL_SOLVER chol_solver= (CHOL_SOLVER)NULL, TYPE_MATRIX type_matrix= (TYPE_MATRIX)NULL, 
		    TYPE_ORIENTATION type_orientation= ORIENTED);

    // symbolic_fact_MMt() is overloaded function
    // this is for GENERAL matrices
    void symbolic_fact_MMt(int m, int n, int nz, int *icola, int *inirowa, double *a, 
		      int k=1, int *prov_pfa= NULL, int *prov_ipfa= NULL, 
		      NUMBERING prov_pfa_numbering= NOT_COMPUTED); 
    // this is for NETWORK matrices (oriented and nonoriented)
    void symbolic_fact_MMt(int nnu, int nar, int *src, int *dst, int k=1, 
		      int *prov_pfa= NULL, int *prov_ipfa= NULL, 
		      NUMBERING prov_pfa_numbering_= NOT_COMPUTED);
    // this is for IDENTITY or IDTY_IDTY matrices
    void symbolic_fact_MMt(int m, int k=1);
    // this is for DIAGONAL matrices
    void symbolic_fact_MMt(int m, double *d1_in, int k=1);
    // this is for DIAG_DIAG matrices
    void symbolic_fact_MMt(int m, double *d1_in, double *d2_in, int k=1);
    // this is for DENSE_DIAG_DIAG matrices
    void symbolic_fact_MMt(int m, int nd, int num_blocks, double *ad, bool d1exist, bool d2exist, double *d1d, double *d2d, double cd1, double cd2);

    void numeric_fact_MMt(double *Theta, int i_k=0);
    void numeric_solve_MMt(double *rhs, int i_k= 0, WHO_PERMUTES whoperm= (WHO_PERMUTES)NULL);

    // routines for positive semidefinite matrices M (not M*M')
    // symbolic_fact_M() is overloaded function
    // this is for GEN_SYM_UPTR matrices
    void symbolic_fact_M(int m, int nz, int *icola, int *inirowa, 
			 int *prov_pfa= NULL, int *prov_ipfa= NULL, 
			 NUMBERING prov_pfa_numbering= NOT_COMPUTED); 
    // this is for DIAGONAL matrices
    void symbolic_fact_M(int m);
    // this is for TRIDIAG_SHIFTED matrices
    void symbolic_fact_M(int m, int t_shift);

    void numeric_fact_M(double *a, double *b=NULL);
    void numeric_solve_M(double *rhs, WHO_PERMUTES whoperm= (WHO_PERMUTES)NULL);

    // access to data in class
    int get_pfa(int i); // return pfa[i] with no check 
    int get_ipfa(int i); // return ipfa[i] with no check
    int get_maxlnz(); // return maxlnz with no check
    int get_maxfillin(); // return maxfillin with no check
    int get_njka(); // return njka with no check
    int get_num_zero_pivots(); // returns num_zero_pivots with no check
    int get_num_semidef_matrix(); // returns num_semidef_matrix with no check

private:

    void free_mem();

    void symbolic_fact_MMt_sprsblkllt_general(int n, int nz, int *icola, int *inirowa, double *a );
    void symbolic_fact_MMt_sprsblkllt_network(int *src, int *dst);

    void get_indices_a_general(int *icola, int *inirowa, double *a); 
    static bool comp_field1(const SRC_DST_ARC & i, const SRC_DST_ARC & j);
    static bool comp_field2(const SRC_DST_ARC & i, const SRC_DST_ARC & j);
    void get_ipk_ipl_network(int *src, int *dst); 
    void get_pfa_ipfa_network();
    void get_pfa_ipfa_general(int *inp_ia, int *inp_ja);
    void symbolic_AThetaAt_A(); //common to general and network matrices A*A' and symmetric matrices A
    void get_ilnz_network();
    void get_ilnz_ifillin_general(int *inp_ia, int *inp_ja);

    void numeric_fact_MMt_sprsblkllt_general(double *Theta, int i_k);
    void numeric_fact_MMt_sprsblkllt_network(double *Theta, int i_k);
    void numeric_fact_MMt_identity(double *Theta, int i_k);
    void numeric_fact_MMt_idty_idty(double *Theta, int i_k);
    void numeric_fact_MMt_diagonal(double *Theta, int i_k);
    void numeric_fact_MMt_diag_diag(double *Theta, int i_k);
    void numeric_fact_MMt_dense_diag_diag(double *Theta, int i_k);

    void numeric_solve_MMt_sprsblkllt(double *rhs, int i_k, WHO_PERMUTES whoperm); //common to general and network matrices
    void numeric_solve_MMt_diag(double *rhs, int i_k);
    void numeric_solve_MMt_dense(double *rhs, int i_k); 

    void symbolic_fact_M_sprsblkllt_general(int nz, int *icola, int *inirowa);

    void numeric_fact_M_sprsblkllt_general(double *a);
    void numeric_fact_M_diagonal(double *d1);
    void numeric_fact_M_tridiag_shifted(double *tridiag_d, double *tridiag_s);

    void numeric_solve_M_sprsblkllt_general(double *rhs, WHO_PERMUTES whoperm); 
    void numeric_solve_M_diagonal(double *rhs);
    void numeric_solve_M_tridiag_shifted(double *rhs);
};

inline void SparseChol::symbolic_fact_MMt(int m, int n, int nz, int *icola, int *inirowa, 
				  double *a, int k, int *prov_pfa, int *prov_ipfa, NUMBERING prov_pfa_numbering)
//!  Computes symmetric ordering and symbolic factorization of AA' for a general matrix A.
/*!  A is general matrix in compressed rowwise order.
   
  @param m Number of rows of A.
  @param n Number of columns of A.
  @param nz Number of nonzeros in A.
  @param icola(nz) Column of each element of A.
  @param inirowa(m+1) Pointers to first by row in icola().
  @param a(nz) Matrix elements.
  @param k Number of matrices with same topology to be factorized.
  @param prov_pfa(m),prov_ipfa(m) Ordering provided by user, to be used; 
  the routine then only performs the symbolic factorization using
  this ordering. If NULL, then no ordering is provided and
  it will be computed. NULL is the default value if the
  parameter not provided. No check performed about the
  correctness of the ordering provided.
  @param prov_pfa_numbering Numbering of provided ordering.
*/
{
    this->m= m;
    this->k= k;
/*
  // not yet implemented: the computation of pfa and ipfa computes other 
  // needed vectors, so it is not simple

  if ((pfa != NULL) && (ipfa != NULL)) {
  for(int i=0; i<m; i++) {
  this->pfa[i]= pfa[i];
  }
  for(int i=0; i<m; i++) {
  this->ipfa[i]= ipfa[i];
  }
  }
*/
  switch(chol_solver) {
    case SPRSBLKLLT:
      symbolic_fact_MMt_sprsblkllt_general(n,nz,icola,inirowa,a);
      break;
    case CHOLMOD:
      throw ExceptionBlockIP(CHOLMOD_NOT_IMPLEMENTED, __FILE__, __LINE__);
      break;
  }
}

inline void SparseChol::symbolic_fact_MMt(int nnu, int nar, int *src, int *dst, int k, int *prov_pfa, int *prov_ipfa, NUMBERING prov_pfa_numbering)
//!  Computes symmetric ordering and symbolic factorization of AA' for a network matrix A.

/*!  A is node-arc incidence matrix (last node not considered)

  @param nnu Number of nodes.
  @param nar Number of arcs.
  @param src(nar)  Source node for each arc. 
  @param dst(nar)  Destination node for each arc.
  @param k Number of matrices with same topology to be factorized.
  @param prov_pfa(m),prov_ipfa(m) Ordering provided by user, to be used; 
  the routine then only performs the symbolic factorization using
  this ordering. If NULL, then no ordering is provided and
  it will be computed. NULL is the default value if the
  parameter not provided. No check performed about the
  correctness of the ordering provided.
  @param prov_pfa_numbering  Numbering of provided ordering.
*/
{
  this->nnu= nnu;
  this->nar= nar;
  m= nnu-1; //last node removed
  this->k= k;
/*
  // not yet implemented: the computation of pfa and ipfa computes other 
  // needed vectors, so it is not simple

  if ((pfa_ != NULL) && (ipfa_ != NULL)) {
    for(int i=0; i<m; i++) {
      pfa[i]= pfa_[i];
    }
    for(int i=0; i<m; i++) {
      ipfa[i]= ipfa_[i];
    }
  }
*/
  switch(chol_solver) {
    case SPRSBLKLLT:
      symbolic_fact_MMt_sprsblkllt_network(src,dst);
      break;
    case CHOLMOD:
      throw ExceptionBlockIP(CHOLMOD_NOT_IMPLEMENTED, __FILE__, __LINE__);
      break;
  }
}

inline void SparseChol::numeric_fact_MMt(double *Theta, int i_k)
//!  Computes numerical factorization of A*Theta*A'.
/*!
  @param Theta(n) Diagonal matrix Theta of A*Theta*A', of size number of columns of A.
  @param i_k Which matrix to factorize, from 0 to k-1.
  @note symbolic_fact_MMt() must have been previously called.
*/
{
  switch (type_matrix) {
    // these type of matrices involve a diagonal M*Mt matrix
  case IDENTITY:
    numeric_fact_MMt_identity(Theta,i_k);
    break;
  case IDTY_IDTY:
    numeric_fact_MMt_idty_idty(Theta,i_k);
    break;
  case DIAGONAL:
    numeric_fact_MMt_diagonal(Theta,i_k);
    break;
  case DIAG_DIAG:
    numeric_fact_MMt_diag_diag(Theta,i_k);
    break;

    // this type of matrices requires a dense Cholesky solver 
  case DENSE_DIAG_DIAG:
    numeric_fact_MMt_dense_diag_diag(Theta,i_k);    
    break;
   
    // these type of matrices need a sparse Cholesky solver
  case GENERAL:
    switch(chol_solver) {
    case SPRSBLKLLT:
      numeric_fact_MMt_sprsblkllt_general(Theta,i_k);
      break;
    case CHOLMOD:
      throw ExceptionBlockIP(CHOLMOD_NOT_IMPLEMENTED, __FILE__, __LINE__);
      break;
    }
    break;
  case NETWORK:
    switch(chol_solver) {
    case SPRSBLKLLT:
      numeric_fact_MMt_sprsblkllt_network(Theta,i_k);
      break;
    case CHOLMOD:
      throw ExceptionBlockIP(CHOLMOD_NOT_IMPLEMENTED, __FILE__, __LINE__);
      break;
    }
    break;
  default:
    throw ExceptionBlockIP(NOT_FOR_THIS_MATRIX_TYPE, __FILE__, __LINE__);
  }
}

inline void SparseChol::numeric_solve_MMt(double *rhs, int i_k, WHO_PERMUTES whoperm)
//! Solve systems with matrix A*Theta*A' and right-hand-side rhs.
/*! The solution is returned by the same vector rhs (then it is overwritten).
   
  @param rhs On input vector rhs; on output the solution vector.
  @param whoperm  If CHOLESKY, then the reordering by pfa/ipfa must internally be 
  applied to rhs by the routine;
  if USER_OF_CHOLESKY, then the user already provides rhs in the 
  permutted order given by pfa/ipfa.

  @param i_k Which matrix to use, from 0 to k-1.

  @note symbolic_fact_MMt() and numeric_fact_MMt() must have been previously called.
*/
{
  switch (type_matrix) {
    // these type of matrices involve a diagonal M*Mt matrix
  case IDENTITY:
  case IDTY_IDTY:
  case DIAGONAL:
  case DIAG_DIAG:
    numeric_solve_MMt_diag(rhs,i_k);
    break;
    
    // this type of matrices requires a dense Cholesky solver 
  case DENSE_DIAG_DIAG:
    numeric_solve_MMt_dense(rhs,i_k);    
    break;

   // these type of matrices need a sparse Cholesky solver
  case GENERAL:
  case NETWORK:
    switch(chol_solver) {
    case SPRSBLKLLT:
      numeric_solve_MMt_sprsblkllt(rhs,i_k,whoperm);
      break;
    case CHOLMOD:
      throw ExceptionBlockIP(CHOLMOD_NOT_IMPLEMENTED, __FILE__, __LINE__);
      break;
    }
    break;
  default:
    throw ExceptionBlockIP(NOT_FOR_THIS_MATRIX_TYPE, __FILE__, __LINE__);
  }
}

inline void SparseChol::symbolic_fact_M(int m, int nz, int *icola, int *inirowa, 
					int *prov_pfa, int *prov_ipfa, NUMBERING prov_pfa_numbering)
//!  Computes symmetric ordering and symbolic factorization of A for a general symmetric matrix
/*!  A is general symmetric upper triangular (with diagonal) matrix in compressed rowwise order
   
  @param m Number of rows/columns of A.
  @param nz Number of nonzeros in A.
  @param icola(nz) Column of each element of A (only diagonal and upper triangle)
  @param inirowa(m+1) Pointers to first by row in icola().
  @param prov_pfa(m),prov_ipfa(m) Ordering provided by user, to be used; 
  the routine then only performs the symbolic factorization using
  this ordering. If NULL, then no ordering is provided and
  it will be computed. NULL is the default value if the
  parameter not provided. No check performed about the
  correctness of the ordering provided.
  @param prov_pfa_numbering Numbering of provided ordering.
*/
{
    this->m= m;
/*
  // not yet implemented: the computation of pfa and ipfa computes other 
  // needed vectors, so it is not simple

  if ((pfa != NULL) && (ipfa != NULL)) {
  for(int i=0; i<m; i++) {
  this->pfa[i]= pfa[i];
  }
  for(int i=0; i<m; i++) {
  this->ipfa[i]= ipfa[i];
  }
  }
*/
  switch(chol_solver) {
    case SPRSBLKLLT:
      symbolic_fact_M_sprsblkllt_general(nz,icola,inirowa);
      break;
    case CHOLMOD:
      throw ExceptionBlockIP(CHOLMOD_NOT_IMPLEMENTED, __FILE__, __LINE__);
      break;
  }
}

inline void SparseChol::numeric_fact_M(double *a, double*b)
//!  Computes numerical factorization for a positive semidefinite matrix A
/*!  A is either diagonal or general symmetric upper triangular (with diagonal) matrix in compressed rowwise order

  @param a(nza) First vector of matrix elements (if general matrix, then nza= nz and a() 
      has diagonal and upper triangle; if diagonal matrix, then nza=m and a() has the diagonal;
      if tridiagonal shifted matrix, then nza=m and a() has the main diagonal of the matrix)
  @param b(nzb) Second vector (if needed) of matrix elements (if tridigional shifted matrix, then nzb=m+m-t_shifted
      and a() has the non-main diagonal; for other type of matrices b=NULL)
  @note symbolic_fact_M() must have been previously called.
*/
{
  switch (type_matrix) {
  case DIAGONAL:
    numeric_fact_M_diagonal(a);
    break;
  case GEN_SYM_UPTR:
    switch(chol_solver) {
    case SPRSBLKLLT:
      numeric_fact_M_sprsblkllt_general(a);
      break;
    case CHOLMOD:
      throw ExceptionBlockIP(CHOLMOD_NOT_IMPLEMENTED, __FILE__, __LINE__);
      break;
    }
    break;
  case TRIDIAG_SHIFTED:
    numeric_fact_M_tridiag_shifted(a,b);
    break;
  }
}

inline void SparseChol::numeric_solve_M(double *rhs, WHO_PERMUTES whoperm)
//! Solve systems with a positive semidefinite symmetric matrix A and right-hand-side rhs.
/*! The solution is returned by the same vector rhs (then it is overwritten).
   
  @param rhs On input vector rhs; on output the solution vector.
  @param whoperm  If CHOLESKY, then the reordering by pfa/ipfa must internally be 
  applied to rhs by the routine;
  if USER_OF_CHOLESKY, then the user already provides rhs in the 
  permutted order given by pfa/ipfa.

  @note symbolic_fact_M() and numeric_fact_M() must have been previously called.
*/
{ 
  switch (type_matrix) {
  case DIAGONAL:
    numeric_solve_M_diagonal(rhs);
    break;
  case IDENTITY:
    // nothing to be done: rhs on ouput is rhs on input
    break;
  case GEN_SYM_UPTR:
    switch(chol_solver) {
    case SPRSBLKLLT:
      numeric_solve_M_sprsblkllt_general(rhs,whoperm);
      break;
    case CHOLMOD:
      throw ExceptionBlockIP(CHOLMOD_NOT_IMPLEMENTED, __FILE__, __LINE__);
      break;
    }
    break;
  case TRIDIAG_SHIFTED:
    numeric_solve_M_tridiag_shifted(rhs);
    break;
  default:
    throw ExceptionBlockIP(NOT_FOR_THIS_MATRIX_TYPE, __FILE__, __LINE__);
  }
}

inline int SparseChol::get_pfa(int i) 
//! Provides pfa[i] with no check.
/*! 
  @return pfa[i]
  @note The user must guarantee pfa previously computed in call to symbolic_fact_MMt().
 */
{
    return(pfa[i]);
}

inline int SparseChol::get_ipfa(int i)
//! Provides ipfa[i] with no check.
/*! 
  @return ipfa[i]
  @note The user must guarantee ipfa previously computed in call to symbolic_fact_MMt().
 */
{
    return(ipfa[i]);
}

inline int SparseChol::get_maxlnz()
//! Provides maxlnz with no check.
/*! 
  @return maxlnz
  @note The user must guarantee maxlnz previously computed in call to symbolic_fact_MMt().
 */
{
    return (maxlnz);
}

inline int SparseChol::get_maxfillin()
//! Provides maxfillin with no check.
/*! 
  return maxfillin
  @note The user must guarantee maxfillin previously computed in call to symbolic_fact_MMt().
 */
{
    return (maxfillin);
}

inline int SparseChol::get_njka()
//! Provides njka with no check.
/*! 
    @return njka
  @note The user must guarantee njka previously computed in call to symbolic_fact_MMt().
 */
{
    return (njka);
}

inline int SparseChol::get_num_zero_pivots()
//! Provides num_zero_pivots with no check.
/*! 
  @return num_zero_pivots
  @note It will 0 if no numeric factorization made.
 */
{
    return (num_zero_pivots);
}

inline int SparseChol::get_num_semidef_matrix()
//! Provides num_semidef_matrix with no check.
/*! 
  @return num_semidef_matrix
  @note It will 0 if no numeric factorization made.
 */
{
    return (num_semidef_matrix);
}

#endif //SPCHOL_H
